from abc import abstractmethod, ABCMeta

import numpy as np
from numpy.linalg import lstsq
import scipy.stats

from sureal.core.mixin import TypeVersionEnabled


class PerfMetric(TypeVersionEnabled):

    __metaclass__ = ABCMeta

    @classmethod
    @abstractmethod
    def _preprocess(cls, groundtruths, predictions, **kwargs):
        raise NotImplementedError

    @classmethod
    @abstractmethod
    def _evaluate(cls, groundtruths, predictions, **kwargs):
        raise NotImplementedError

    def __init__(self, groundtruths, predictions):
        """
        Performance metric on quality metrics
        :param groundtruths: either list of real numbers (aggregate scores like
        MOS or DMOS or MLE), or list of list of real numbers (list of raw scores)
        :param predictions: list of real numbers
        :return:
        """
        TypeVersionEnabled.__init__(self)
        self.groundtruths = groundtruths
        self.predictions = predictions
        self._assert_args()

    def _assert_args(self):
        assert len(self.groundtruths) == len(self.predictions), 'The lengths of groundtruth labels and predictions do not match.'

    def evaluate(self, **kwargs):
        """
        :return: ret - a dictionary with 'score' and other keys
        """
        groundtruths, predictions = self._preprocess(self.groundtruths, self.predictions, **kwargs)
        result = self._evaluate(groundtruths, predictions, **kwargs)
        assert 'score' in result, 'Score does not exist in result.'
        return result


class AggrScorePerfMetric(PerfMetric):
    """
    Groundtruth is a list of aggregate scores (list of real numbers)
    """

    @staticmethod
    def sigmoid_adjust(xs, ys):
        ys_max = np.max(ys) + 0.1
        ys_min = np.min(ys) - 0.1

        # normalize to [0, 1]
        ys = list((np.array(ys) - ys_min) / (ys_max - ys_min))

        zs = -np.log(1.0 / np.array(ys).T - 1.0)
        Y_mtx = np.array((np.ones(len(ys)), zs)).T
        x_vec = np.array(xs).T
        a_b = lstsq(Y_mtx, x_vec, rcond=1)[0]
        a = a_b.item(0)
        b = a_b.item(1)

        xs = 1.0 / (1.0 + np.exp(- (np.array(xs) - a) / b))

        # denormalize
        xs = xs * (ys_max - ys_min) + ys_min

        return xs

    @classmethod
    def _preprocess(cls, groundtruths, predictions, **kwargs):
        aggre_method = kwargs['aggr_method'] if 'aggr_method' in kwargs else np.mean
        enable_mapping = kwargs['enable_mapping'] if 'enable_mapping' in kwargs else False

        groundtruths_ = list(map(
            lambda x: aggre_method(x) if hasattr(x, '__len__') else x,
            groundtruths))

        if enable_mapping:
            predictions_ = cls.sigmoid_adjust(predictions, groundtruths_)
        else:
            predictions_ = predictions

        return groundtruths_, predictions_


class RmsePerfMetric(AggrScorePerfMetric):

    TYPE = "RMSE"
    VERSION = "1.0"

    @classmethod
    def _evaluate(cls, groundtruths, predictions, **kwargs):
        rmse = np.sqrt(np.mean(np.power(np.array(groundtruths) - np.array(predictions), 2.0)))
        result = {'score': rmse}
        return result


class SrccPerfMetric(AggrScorePerfMetric):

    TYPE = "SRCC"
    VERSION = "1.0"

    @classmethod
    def _evaluate(cls, groundtruths, predictions, **kwargs):
        # spearman
        srcc, _ = scipy.stats.spearmanr(groundtruths, predictions)
        result = {'score': srcc}
        return result


class PccPerfMetric(AggrScorePerfMetric):

    TYPE = "PCC"
    VERSION = "1.0"

    @classmethod
    def _evaluate(cls, groundtruths, predictions, **kwargs):
        # pearson
        pcc, _ = scipy.stats.pearsonr(groundtruths, predictions)
        result = {'score': pcc}
        return result


class KendallPerfMetric(AggrScorePerfMetric):

    TYPE = "KENDALL"
    VERSION = "1.0"

    @classmethod
    def _evaluate(cls, groundtruths, predictions, **kwargs):
        # kendall
        kendall, _ = scipy.stats.kendalltau(groundtruths, predictions)
        result = {'score': kendall}
        return result



import copy

import numpy as np

from sureal.dataset_reader import SelectSubjectRawDatasetReader, SelectDisVideoRawDatasetReader
from sureal.subjective_model import MaximumLikelihoodEstimationModel


class MaximumLikelihoodEstimationModelWithBootstrapping(MaximumLikelihoodEstimationModel):

    TYPE = 'MLE_BSTP'
    VERSION = MaximumLikelihoodEstimationModel.VERSION + "_0.1"

    DEFAULT_N_BOOTSTRAP = 30
    DEFAULT_BOOTSTRAP_SUBJECTS = True
    DEFAULT_BOOSTRAP_DIS_VIDEOS = True

    @classmethod
    def _run_modeling(cls, dataset_reader, **kwargs):

        force_subjbias_zeromean = kwargs['force_subjbias_zeromean'] \
            if 'force_subjbias_zeromean' in kwargs \
               and kwargs['force_subjbias_zeromean'] is not None else MaximumLikelihoodEstimationModel.DEFAULT_FORCE_SUBJBIAS_ZEROMEAN
        assert isinstance(force_subjbias_zeromean, bool)
        new_kwargs = copy.deepcopy(kwargs)
        new_kwargs['force_subjbias_zeromean'] = False

        result = super(MaximumLikelihoodEstimationModelWithBootstrapping, cls).\
            _run_modeling(dataset_reader, **new_kwargs)

        dataset = dataset_reader.to_dataset()
        n_subj = dataset_reader.num_observers
        n_disvideo = dataset_reader.num_dis_videos

        n_bootstrap = kwargs['n_bootstrap'] if 'n_bootstrap' in kwargs \
                                               and kwargs['n_bootstrap'] is not None else cls.DEFAULT_N_BOOTSTRAP
        assert isinstance(n_bootstrap, int) and n_bootstrap > 0

        bootstrap_subjects = kwargs['bootstrap_subjects'] \
            if 'bootstrap_subjects' in kwargs \
               and kwargs['bootstrap_subjects'] is not None else cls.DEFAULT_BOOTSTRAP_SUBJECTS
        assert isinstance(bootstrap_subjects, bool)

        boostrap_dis_videos = kwargs['boostrap_dis_videos'] \
            if 'boostrap_dis_videos' in kwargs \
               and kwargs['boostrap_dis_videos'] is not None else cls.DEFAULT_BOOSTRAP_DIS_VIDEOS
        assert isinstance(boostrap_dis_videos, bool)

        if bootstrap_subjects:
            quality_scores_ci95 = \
                cls._bootstrap_subjects(dataset, result, n_subj, n_bootstrap, new_kwargs)
            result['quality_scores_ci95'] = quality_scores_ci95
        else:
            del result['quality_scores_ci95']

        if boostrap_dis_videos:
            observer_bias_ci95, observer_inconsistency_ci95 = \
                cls._boostrap_dis_videos(dataset, result, n_disvideo, n_bootstrap, new_kwargs)
            result['observer_bias_ci95'] = observer_bias_ci95
            result['observer_inconsistency_ci95'] = observer_inconsistency_ci95
        else:
            del result['observer_bias_ci95']
            del result['observer_inconsistency_ci95']

        if force_subjbias_zeromean is True:
            assert 'quality_scores' in result
            assert 'observer_bias' in result
            mean_b_s = np.mean(result['observer_bias'])
            result['observer_bias'] = list(np.array(result['observer_bias']) - mean_b_s)
            result['quality_scores'] = list(np.array(result['quality_scores']) + mean_b_s)

        return result

    @classmethod
    def _bootstrap_subjects(cls, dataset, result, n_subj, n_bootstrap, kwargs):
        bootstrap_results = []
        for ibootstrap in range(n_bootstrap):
            print(f"Bootstrap with seed {ibootstrap}")

            np.random.seed(ibootstrap)
            selected_subjects = np.random.choice(range(n_subj), size=n_subj, replace=True)

            select_subj_reader = SelectSubjectRawDatasetReader(
                dataset, input_dict={'selected_subjects': selected_subjects})

            bootstrap_result = super(MaximumLikelihoodEstimationModelWithBootstrapping, cls). \
                _run_modeling(select_subj_reader, **kwargs)

            bootstrap_observer_bias_offset = np.mean(
                np.array(bootstrap_result['observer_bias']) -
                np.array(result['observer_bias'])[selected_subjects]
            )

            bootstrap_result['observer_bias'] = list(np.array(bootstrap_result['observer_bias']) -
                                                     bootstrap_observer_bias_offset)
            bootstrap_result['quality_scores'] = list(np.array(bootstrap_result['quality_scores']) +
                                                      bootstrap_observer_bias_offset)

            bootstrap_results.append(bootstrap_result)
        bootstrap_quality_scoress = np.array([r['quality_scores'] for r in bootstrap_results])
        quality_scores_ci95 = [
            np.array(result['quality_scores']) - np.percentile(bootstrap_quality_scoress, 2.5, axis=0),
            np.percentile(bootstrap_quality_scoress, 97.5, axis=0) - np.array(result['quality_scores'])
        ]
        return quality_scores_ci95

    @classmethod
    def _boostrap_dis_videos(cls, dataset, result, n_dis_videos, n_bootstrap, kwargs):
        bootstrap_results = []
        for ibootstrap in range(n_bootstrap):
            print(f"Bootstrap with seed {ibootstrap}")

            np.random.seed(ibootstrap)
            selected_dis_videos = np.random.choice(range(n_dis_videos), size=n_dis_videos, replace=True)

            select_dis_video_reader = SelectDisVideoRawDatasetReader(
                dataset, input_dict={'selected_dis_videos': selected_dis_videos})

            bootstrap_result = super(MaximumLikelihoodEstimationModelWithBootstrapping, cls). \
                _run_modeling(select_dis_video_reader, **kwargs)

            bootstrap_quality_scores_offset = np.mean(
                np.array(bootstrap_result['quality_scores']) -
                np.array(result['quality_scores'])[selected_dis_videos]
            )

            bootstrap_result['quality_scores'] = list(np.array(bootstrap_result['quality_scores']) -
                                                      bootstrap_quality_scores_offset)
            bootstrap_result['observer_bias'] = list(np.array(bootstrap_result['observer_bias']) +
                                                     bootstrap_quality_scores_offset)

            bootstrap_results.append(bootstrap_result)
        bootstrap_observer_biass = np.array([r['observer_bias'] for r in bootstrap_results])
        observer_bias_ci95 = [
            np.array(result['observer_bias']) - np.percentile(bootstrap_observer_biass, 2.5, axis=0),
            np.percentile(bootstrap_observer_biass, 97.5, axis=0) - np.array(result['observer_bias'])
        ]
        bootstrap_observer_inconsistencys = np.array([r['observer_inconsistency'] for r in bootstrap_results])
        observer_inconsistency_ci95 = [
            np.array(result['observer_inconsistency']) - np.percentile(bootstrap_observer_inconsistencys, 2.5, axis=0),
            np.percentile(bootstrap_observer_inconsistencys, 97.5, axis=0) - np.array(result['observer_inconsistency'])
        ]
        return observer_bias_ci95, observer_inconsistency_ci95


class MaximumLikelihoodEstimationModelContentObliviousWithBootstrapping(MaximumLikelihoodEstimationModelWithBootstrapping):
    TYPE = 'MLE_CO_BSTP' # maximum likelihood estimation (no content modeling) with bootstrapping
    VERSION = MaximumLikelihoodEstimationModelWithBootstrapping.VERSION + "_0.1"
    mode = 'CONTENT_OBLIVIOUS'


import sys
import time
from functools import partial

import numpy as np
from scipy import linalg
from scipy.optimize import minimize
from scipy.stats import norm

from sureal.subjective_model import SubjectiveModel
from sureal.dataset_reader import PairedCompDatasetReader

__copyright__ = "Copyright 2016-2019, Netflix, Inc."
__license__ = "Apache, Version 2.0"


class PairedCompSubjectiveModel(SubjectiveModel):

    def _assert_args(self):
        super(PairedCompSubjectiveModel, self)._assert_args()
        assert isinstance(self.dataset_reader, PairedCompDatasetReader)

    @classmethod
    def from_dataset_file(cls, dataset_filepath, content_ids=None, asset_ids=None):
        dataset = cls._import_dataset_and_filter(dataset_filepath, content_ids, asset_ids)
        dataset_reader = PairedCompDatasetReader(dataset)
        return cls(dataset_reader)

    @staticmethod
    def _get_ref_mos(dataset_reader, mos):
        raise NotImplementedError

    @staticmethod
    def _get_opinion_score_3darray_with_preprocessing(dataset_reader, **kwargs):
        raise NotImplementedError


class BradleyTerryNewtonRaphsonPairedCompSubjectiveModel(PairedCompSubjectiveModel):
    """ Bradley-Terry model to convert paired comparison scores to continuous score.

    Implementation based on: http://sites.stat.psu.edu/~drh20/code/btmatlab/btnr.m

    function info=btnr(wm)
    % function info=btnr(wm)
    %
    % btnr uses a Newton-Raphson algorithm to fit the Bradley-Terry model.
    %
    % The input wm is an nxn matrix such that wm(i,j) is the number of times
    % team i beat team j.
    %
    % The output is an nx1 vector of parameter estimates of "team skill".
    %
    % This algorithm does not contain any checks for lack of convergence;
    % it is assumed that for any partition of the teams into sets A and B,
    % at least one team from A beats at least one team from B at least once.

    n=size(wm,1);
    %flops(0);

    % The parameter vector gamma is constrained to keep its last
    % entry equal to 0; thus, nmo stands for "n minus one".
    nmo=n-1;
    gamma=zeros(n,1); % initial value:  All teams equal
    iteration=0;
    change=realmax;
    gm=wm(1:nmo,:)+wm(:,1:nmo)';
    wins=sum(wm(1:nmo,:),2);
    gind=(gm>0);
    squaregind=gind(:,1:nmo);
    y=zeros(nmo,n);
    rus=y;

    while stats.norm(change) > 1e-08
        iteration=iteration+1;
        pi=exp(gamma);
        pius=pi(:,ones(n,1));
        piust=pius(:,1:nmo)';
        pius=pius(1:nmo,:);
        rus(gind)=pius(gind) ./ (pius(gind)+piust(gind));
        y(gind) = gm(gind) .* rus(gind);
        d2Q=-sum(y,2);
        dL=wins+d2Q;
        y(gind)=y(gind).*(1-rus(gind));
        d2L=-diag(sum(y,2));
        d2L(squaregind)=y(squaregind);
        A=d2L;

        cov = -inv(d2L);

        change=(A) \\ dL;
        %change = inv(d2L) * dL;
        change(n) = 0;
        gamma = gamma - change;
    end
    Iterations = iteration
    %Floating_point_operations=flops
    info=exp(gamma);
    info(n)=1;

    """

    TYPE = 'BT_NR'
    VERSION = '1.0'

    @classmethod
    def _run_modeling(cls, dataset_reader, **kwargs):

        wm = np.nansum(dataset_reader.opinion_score_3darray, axis=2)

        # wm = np.array(
        #     [[0, 3, 2, 7],
        #      [1, 0, 6, 3],
        #      [4, 3, 0, 0],
        #      [1, 2, 5, 0]]
        #               )

        n, m = wm.shape
        assert n == m

        nmo = n - 1
        gamma = np.zeros(n)
        iteration = 0
        change = sys.float_info.max
        gm = wm[0:nmo, :] + wm[:, 0:nmo].transpose()
        wins = np.sum(wm[0:nmo, :], axis=1)
        gind = (gm > 0)
        squaregind = gind[:, 0:nmo]
        y = np.zeros([nmo, n])
        rus = y.copy()
        cov = None

        DELTA_THR = 1e-8

        while linalg.norm(change) > DELTA_THR:
            iteration += 1
            pi = np.exp(gamma)
            pius = np.tile(pi, (n, 1)).transpose()
            piust = pius[:, 0:nmo].transpose()
            pius = pius[0:nmo, :]
            rus[gind] = np.divide(pius[gind], (pius[gind] + piust[gind]))
            y[gind] = np.multiply(gm[gind], rus[gind])
            d2Q = - np.sum(y, axis=1)
            dL = wins + d2Q
            y[gind] = np.multiply(y[gind], (1. - rus[gind]))
            d2L = -np.diag(np.sum(y, axis=1))
            d2L[squaregind] = y[:, :-1][squaregind]

            # cov = - linalg.pinv(d2L)

            change = np.matmul(linalg.pinv(d2L), dL)
            change = np.hstack([change, np.array([0])])
            gamma -= change

            msg = 'Iteration {itr:4d}: change {change}, mean x_e {x_e}'.format(itr=iteration, change=linalg.norm(change), x_e=np.mean(gamma))
            sys.stdout.write(msg + '\r')
            sys.stdout.flush()
            time.sleep(0.1)

        # scores = np.exp(gamma)
        # scores[-1] = 1.
        # instead of original formulation, output non-exponential score
        scores = gamma
        scores[-1] = 0.0

        # std = np.diagonal(cov)
        # std = np.hstack([std, np.array([0])])

        zscore_output = kwargs['zscore_output'] \
            if 'zscore_output' in kwargs and kwargs['zscore_output'] is not None else False

        if zscore_output:
            scores_mean = np.mean(scores)
            scores_std = np.std(scores)
            scores = (scores - scores_mean) / scores_std
            # std = std / scores_std

        result = {'quality_scores': list(scores),
                  'quality_scores_std': None}
        return result


class BradleyTerryMlePairedCompSubjectiveModel(PairedCompSubjectiveModel):
    """ Bradley-Terry model based on maximum likelihood estimation, classical version.
    """

    TYPE = 'BT_MLE'
    VERSION = '1.0'

    @classmethod
    def _run_modeling(cls, dataset_reader, **kwargs):

        alpha = np.nansum(dataset_reader.opinion_score_3darray, axis=2)

        v, stdv_v, p, stdv_p, cova_v, cova_p = cls.resolve_model(alpha, **kwargs)

        return {'quality_scores': v,
                'quality_scores_std': stdv_v,
                'quality_scores_ci95': [list(1.95996 * np.array(stdv_v)), list(1.95996 * np.array(stdv_v))],
                'quality_scores_p': p,
                'quality_scores_p_std': stdv_p,
                'quality_scores_p_cov': cova_p,
                'quality_scores_v_cov': cova_v}

    @staticmethod
    def resolve_model(alpha, **more):

        display = more['display'] if 'display' in more else True
        assert isinstance(display, bool)

        # example: alpha is paired-comparison matrix
        # alpha = np.array(
        #     [[0, 3, 2, 7],
        #      [1, 0, 6, 3],
        #      [4, 3, 0, 0],
        #      [1, 2, 5, 0]]
        #     )
        M, M_ = alpha.shape
        assert M == M_

        n = alpha + alpha.T

        iteration = 0
        p = 1.0 / M * np.ones(M)
        change = sys.float_info.max

        DELTA_THR = 1e-8

        while change > DELTA_THR:
            iteration += 1
            p_prev = p

            # p_i = (sum_j alpha_ij) / (sum_j n_ij / (p_i + p_j))
            # note that if p = [1, 2, 3, 4] and M = 4, then
            # np.tile(p, (M, 1)).T creates patterns like
            # [
            #  [1, 1, 1, 1],
            #  [2, 2, 2, 2],
            #  [3, 3, 3, 3],
            #  [4, 4, 4, 4]
            #  ]
            pp = np.tile(p, (M, 1)).T + np.tile(p, (M, 1))

            p = np.sum(alpha, axis=1) / np.sum(n / pp, axis=1)  # summing over axis=1 marginalizes j

            p = p / np.sum(p)  # re-normalize to force p a prob. distribution

            change = linalg.norm(p - p_prev)

            if display:
                msg = 'Iteration {itr:4d}: change {change}, mean p {p}'.format(itr=iteration, change=linalg.norm(change), p=np.mean(p))
                sys.stdout.write(msg + '\r')
                sys.stdout.flush()
                time.sleep(0.001)

        # lambda_ii = sum_j -alpha_ij / p_i^2 + n_ij / (p_i + p_j)^2
        # lambda_ij = n_ij / (p_i + p_j)^2, i != j
        # H = [lambda_ij]
        # C = [[-H, 1], [1', 0]]^-1 of (M + 1) x (M + 1)
        # variance of p_i is then diag(C)[i].

        pp = np.tile(p, (M, 1)).T + np.tile(p, (M, 1))
        lbda_ii = np.sum(-alpha / np.tile(p, (M, 1)).T**2 + n / pp**2, axis=1)  # summing over axis=1 marginalizes j
        lbda_ij = n / pp*2
        lbda = lbda_ij + np.diag(lbda_ii)
        cova_p = np.linalg.pinv(
            np.vstack([np.hstack([-lbda, np.ones([M, 1])]), np.hstack([np.ones([1, M]), np.array([[0]])])]))
        vari_p = np.diagonal(cova_p)[:-1]
        stdv_p = np.sqrt(vari_p)
        cova_p = cova_p[:-1, :-1]
        cova_v = cova_p / (np.expand_dims(p, axis=1) * (np.expand_dims(p, axis=1).T))

        v = np.log(p)
        stdv_v = stdv_p / p  # y = log(x) -> dy = 1/x * dx

        return list(v), list(stdv_v), list(p), list(stdv_p), cova_v, cova_p


class ThurstoneMlePairedCompSubjectiveModel(PairedCompSubjectiveModel):
    """ Thurstone model based on maximum likelihood estimation, classical version
    """

    TYPE = 'THURSTONE_MLE'
    # VERSION = '1.0'
    # VERSION = '1.1'  # fix sign nllf sign issue
    VERSION = '1.2'  # add confidence interval

    @classmethod
    def _run_modeling(cls, dataset_reader, **kwargs):

        # example:
        # alpha = np.array(
        #     [[0, 3, 2, 7],
        #      [1, 0, 6, 3],
        #      [4, 3, 0, 0],
        #      [1, 2, 5, 0]]
        #     )
        alpha = np.nansum(dataset_reader.opinion_score_3darray, axis=2)

        scores, std, cov = cls.resolve_model(alpha, **kwargs)

        zscore_output = kwargs['zscore_output'] \
            if 'zscore_output' in kwargs and kwargs['zscore_output'] is not None else False

        if zscore_output:
            scores_mean = np.mean(scores)
            scores_std = np.std(scores)
            scores = (scores - scores_mean) / scores_std
            std = std / scores_std

        result = {
            'quality_scores': scores,
            'quality_scores_std': std,
            'quality_scores_ci95': [list(1.95996 * std), list(1.95996 * std)],
        }
        return result

    @classmethod
    def resolve_model(cls, alpha, **more):

        use_simplified_lbda = more['use_simplified_lbda'] if 'use_simplified_lbda' in more else True
        assert isinstance(use_simplified_lbda, bool)

        M, M_ = alpha.shape
        assert M == M_
        nllf_partial = partial(cls.neg_log_likelihood_function, alpha=alpha)
        v0 = np.zeros(M)
        ret = minimize(nllf_partial, v0, method='SLSQP', jac='2-point',
                       options={'ftol': 1e-8, 'disp': True, 'maxiter': 1000})
        assert ret.success, "minimization is unsuccessful."
        v = ret.x

        if use_simplified_lbda:
            vi_m_vj = np.tile(v, (M, 1)).T - np.tile(v, (M, 1))
            phi_vi_m_vj = norm.cdf(vi_m_vj)
            f_vi_m_vj = norm.pdf(vi_m_vj)
            lbda_ii = np.sum(
                (alpha + alpha.T) * (-vi_m_vj * phi_vi_m_vj * f_vi_m_vj - f_vi_m_vj ** 2) / phi_vi_m_vj ** 2
                , axis=1)  # summing over axis=1 marginalizes j
            lbda_ij = -(
                (alpha + alpha.T) * (-vi_m_vj * phi_vi_m_vj * f_vi_m_vj - f_vi_m_vj ** 2) / phi_vi_m_vj ** 2
            )
        else:
            vi_m_vj = np.tile(v, (M, 1)).T - np.tile(v, (M, 1))
            phi_vi_m_vj = norm.cdf(vi_m_vj)
            f_vi_m_vj   = norm.pdf(vi_m_vj)
            d_vi_m_vj   = - vi_m_vj * norm.pdf(vi_m_vj)
            phi_vj_m_vi = - phi_vi_m_vj
            f_vj_m_vi   =   f_vi_m_vj
            d_vj_m_vi   = - d_vi_m_vj
            lbda_ii = np.sum(
                alpha   * (phi_vi_m_vj * d_vi_m_vj - f_vi_m_vj ** 2) / phi_vi_m_vj ** 2 +
                alpha.T * (phi_vj_m_vi * d_vj_m_vi - f_vj_m_vi ** 2) / phi_vj_m_vi ** 2
            , axis=1)  # summing over axis=1 marginalizes j

            lbda_ij = -(
                alpha   * (phi_vi_m_vj * d_vi_m_vj - f_vi_m_vj ** 2) / phi_vi_m_vj ** 2 +
                alpha.T * (phi_vj_m_vi * d_vj_m_vi - f_vj_m_vi ** 2) / phi_vj_m_vi ** 2
            )

        lbda = lbda_ij + np.diag(lbda_ii)
        cova = np.linalg.pinv(
            np.vstack([np.hstack([-lbda, np.ones([M, 1])]), np.hstack([np.ones([1, M]), np.array([[0]])])]))
        vari = np.diagonal(cova)[:-1]
        stdv = np.sqrt(vari)
        cova = cova[:-1, :-1]

        return v, stdv, cova

    @staticmethod
    def neg_log_likelihood_function(v, alpha):
        # nllf(.) = - sum_i,j log(n_ij / alpha_ij) + alpha_ij * log phi (v_i - v_j) + alpha_ji * log phi (v_j - vi)
        # note that if p = [1, 2, 3, 4] and M = 4, then
        # np.tile(p, (M, 1)).T creates patterns like
        # [
        #  [1, 1, 1, 1],
        #  [2, 2, 2, 2],
        #  [3, 3, 3, 3],
        #  [4, 4, 4, 4]
        #  ]
        M = alpha.shape[0]
        epsilon = 1e-8 / M
        mtx = alpha * np.log(
            norm.cdf(
                (np.tile(v, (M, 1)).T - np.tile(v, (M, 1)))
            ) + epsilon
        ) + alpha.T * np.log(
            norm.cdf(
                (np.tile(v, (M, 1)) - np.tile(v, (M, 1)).T)
            ) + epsilon
        )
        return - np.sum(mtx)



import copy
import math
import os
from typing import Union

import numpy as np
import scipy.stats

from sureal.config import SurealConfig
from sureal.perf_metric import PccPerfMetric, SrccPerfMetric, RmsePerfMetric
from sureal.subjective_model import SubjectiveModel
from sureal.tools.decorator import persist_to_dir
from sureal.tools.stats import histc

try:
    from matplotlib import pyplot as plt

except (ImportError, RuntimeError):
    # This file is sometimes imported too early by __main__.py, before the venv (with matplotlib) is installed
    # OSX system python comes with an ancient matplotlib that triggers RuntimeError when imported in this way
    plt = None

from sureal.dataset_reader import RawDatasetReader, PairedCompDatasetReader, \
    MissingDataRawDatasetReader, SelectSubjectRawDatasetReader
from sureal.tools.misc import import_python_file, import_json_file, Timer, \
    cmap_factory

__copyright__ = "Copyright 2016-2018, Netflix, Inc."
__license__ = "Apache, Version 2.0"


def run_subjective_models(dataset_filepath, subjective_model_classes, do_plot=None, **kwargs):

    def _get_reconstruction_stats(raw_scores, rec_scores):
        assert raw_scores.shape == rec_scores.shape
        rec_scores, raw_scores = zip(*[(rec, raw) for rec, raw in zip(rec_scores.ravel(), raw_scores.ravel())
                                 if (not np.isnan(rec) and not np.isnan(raw))])
        rmse = RmsePerfMetric(raw_scores, rec_scores).evaluate(enable_mapping=False)['score']
        cc = PccPerfMetric(raw_scores, rec_scores).evaluate(enable_mapping=False)['score']
        srocc = SrccPerfMetric(raw_scores, rec_scores).evaluate(enable_mapping=False)['score']
        return {
            'rmse': rmse,
            'cc': cc,
            'srocc': srocc,
        }

    if do_plot is None:
        do_plot = []

    if 'dataset_reader_class' in kwargs:
        dataset_reader_class = kwargs['dataset_reader_class']
    else:
        dataset_reader_class = RawDatasetReader

    if 'dataset_reader_info_dict' in kwargs:
        dataset_reader_info_dict = kwargs['dataset_reader_info_dict']
    else:
        dataset_reader_info_dict = {}

    if 'plot_type' in kwargs:
        plot_type = kwargs['plot_type']
    else:
        plot_type = 'errorbar'

    if 'ax_dict' in kwargs:
        ax_dict = kwargs['ax_dict']
        del kwargs['ax_dict']
    else:
        ax_dict = {}

    raw_score_cmap = kwargs['raw_score_cmap'] if 'raw_score_cmap' in kwargs else cmap_factory('red2green')

    raw_score_residue_range = kwargs['raw_score_residue_range'] if 'raw_score_residue_range' in kwargs else [None, None]
    assert len(raw_score_residue_range) == 2

    colors = ['black', 'gray', 'blue', 'red'] * 2

    if dataset_filepath.endswith('.py'):
        dataset = import_python_file(dataset_filepath)
    elif dataset_filepath.endswith('.json'):
        dataset = import_json_file(dataset_filepath)
    else:
        raise AssertionError("Unknown input type, must be .py or .json")
    dataset_reader = dataset_reader_class(dataset, input_dict=dataset_reader_info_dict)

    subjective_models = [
        s(dataset_reader) for s in subjective_model_classes
    ]

    results = [
        s.run_modeling(**kwargs) for s in subjective_models
    ]

    for result in results:
        dis_video_names = [dis_video['path'] for dis_video in dataset_reader.dis_videos]
        result['dis_video_names'] = dis_video_names

    for subjective_model, result in zip(subjective_models, results):
        if 'raw_scores' in result and 'reconstructions' in result:
            result['reconstruction_stats'] = _get_reconstruction_stats(result['raw_scores'], result['reconstructions'])

    if do_plot == 'all' or 'raw_scores' in do_plot:

        # TODO: visualize repetitions - currently taking mean over repetitions before plotting
        mtx = np.nanmean(dataset_reader.opinion_score_3darray, axis=2).T

        if 'ax_raw_scores' in ax_dict:
            ax_rawscores = ax_dict['ax_raw_scores']
            fig = None
        else:
            n_subj, n_stim = mtx.shape
            w, h = _get_imshow_width_and_height(n_subj, n_stim)
            fig, ax_rawscores = plt.subplots(figsize=(w, h))

        im = ax_rawscores.imshow(mtx, interpolation='nearest', cmap=raw_score_cmap)
        ax_rawscores.set_title(r'Raw Opinion Scores ($u_{ij}$)')
        ax_rawscores.set_xlabel(r'Video Stimuli ($j$)')
        ax_rawscores.set_ylabel(r'Test Subjects ($i$)')
        plt.colorbar(im, ax=ax_rawscores)
        if fig is not None:
            fig.tight_layout()

    if do_plot == 'all' or 'raw_counts' in do_plot:

        # TODO: visualize repetitions - currently taking mean over repetitions before plotting
        mtx = np.nanmean(dataset_reader.opinion_score_3darray, axis=2).T

        if 'ax_raw_counts' in ax_dict:
            ax_rawcounts = ax_dict['ax_raw_counts']
            fig = None
        else:
            n_subj, n_stim = mtx.shape
            w, h = _get_plot_width_and_height(n_stim)
            fig, ax_rawcounts = plt.subplots(figsize=(w, h))

        histcs = np.apply_along_axis(func1d=histc, axis=0, arr=mtx)
        datas = list()
        for ih, h in enumerate(histcs):
            data = [(ih, k, h[k]) for k in sorted(h.keys())]
            datas += data
        xs, ys, vs = zip(*datas)
        ax_rawcounts.scatter(xs, ys, s=np.array(vs) * 2, alpha=0.4)
        ax_rawcounts.set_xlabel(r'Video Stimuli ($j$)')
        ax_rawcounts.set_title(r'Raw Opinion Scores ($u_{ij}$) Counts')
        ax_rawcounts.grid()
        ax_rawcounts.set_xlim(0, len(histcs))
        if fig is not None:
            fig.tight_layout()

    if do_plot == 'all' or 'raw_counts_per_subject' in do_plot:

        # TODO: visualize repetitions - currently taking mean over repetitions before plotting
        mtx = np.nanmean(dataset_reader.opinion_score_3darray, axis=2).T

        if 'raw_counts_per_subject' in ax_dict:
            ax_rawcounts_per_subject = ax_dict['raw_counts_per_subject']
            fig = None
        else:
            n_subj, n_stim = mtx.shape
            w, h = _get_plot_width_and_height(n_subj)
            fig, ax_rawcounts_per_subject = plt.subplots(figsize=(w, h))

        histcs = np.apply_along_axis(func1d=histc, axis=1, arr=mtx)
        datas = list()
        for ih, h in enumerate(histcs):
            data = [(ih, k, h[k]) for k in sorted(h.keys())]
            datas += data
        xs, ys, vs = zip(*datas)
        ax_rawcounts_per_subject.scatter(xs, ys, s=np.array(vs) * 2, alpha=0.4)
        ax_rawcounts_per_subject.set_xlabel(r'Test Subjects ($i$)')
        ax_rawcounts_per_subject.set_title(r'Raw Opinion Scores ($u_{ij}$) Counts')
        ax_rawcounts_per_subject.grid()
        ax_rawcounts_per_subject.set_xlim(0, len(histcs))
        if fig is not None:
            fig.tight_layout()

    if do_plot == 'all' or 'raw_scores_minus_quality_scores' in do_plot:

        for subjective_model, result in zip(subjective_models, results):
            if 'quality_scores' in result:
                quality_scores = result['quality_scores']
                label = subjective_model.TYPE

                # TODO: visualize repetitions - currently taking mean over repetitions before plotting
                mtx = dataset_reader.opinion_score_3darray
                mtx = mtx - np.array(quality_scores)[:, None, None]
                mtx = np.nanmean(mtx, axis=2).T

                if 'ax_raw_scores_minus_quality_scores' in ax_dict:
                    ax_raw_scores_minus_quality_scores = ax_dict['ax_raw_scores_minus_quality_scores']
                    fig = None
                else:
                    w, h = _get_imshow_width_and_height(*mtx.shape)
                    fig, ax_raw_scores_minus_quality_scores = plt.subplots(figsize=(w, h))

                im = ax_raw_scores_minus_quality_scores.imshow(mtx, interpolation='nearest',
                                                               vmin=raw_score_residue_range[0], vmax=raw_score_residue_range[1],
                                                               cmap=raw_score_cmap)
                ax_raw_scores_minus_quality_scores.set_title(r'$u_{ij} - \psi_j$' + ', {}'.format(label))
                ax_raw_scores_minus_quality_scores.set_xlabel(r'Video Stimuli ($j$)')
                ax_raw_scores_minus_quality_scores.set_ylabel(r'Test Subjects ($i$)')
                plt.colorbar(im, ax=ax_raw_scores_minus_quality_scores)
                if fig is not None:
                    fig.tight_layout()

    if do_plot == 'all' or 'raw_scores_minus_quality_scores_and_observer_bias' in do_plot:

        for subjective_model, result in zip(subjective_models, results):
            if 'quality_scores' in result and 'observer_bias' in result:
                observer_bias = result['observer_bias']
                quality_scores = result['quality_scores']
                label = subjective_model.TYPE

                # TODO: visualize repetitions - currently taking mean over repetitions before plotting
                mtx = dataset_reader.opinion_score_3darray
                mtx = mtx - np.array(quality_scores)[:, None, None]
                mtx = mtx - np.array(observer_bias)[None, :, None]
                mtx = np.nanmean(mtx, axis=2).T

                if 'ax_raw_scores_minus_quality_scores_and_observer_bias' in ax_dict:
                    ax_raw_scores_minus_quality_scores_and_observer_bias = ax_dict['ax_raw_scores_minus_quality_scores_and_observer_bias']
                    fig = None
                else:
                    w, h = _get_imshow_width_and_height(*mtx.shape)
                    fig, ax_raw_scores_minus_quality_scores_and_observer_bias = plt.subplots(figsize=(w, h))

                im = ax_raw_scores_minus_quality_scores_and_observer_bias.imshow(mtx, interpolation='nearest',
                                                                                 vmin=raw_score_residue_range[0], vmax=raw_score_residue_range[1],
                                                                                 cmap=raw_score_cmap)
                ax_raw_scores_minus_quality_scores_and_observer_bias.set_title(r'$u_{ij} - \psi_j - \Delta_i$' + ', {}'.format(label))
                ax_raw_scores_minus_quality_scores_and_observer_bias.set_xlabel(r'Video Stimuli ($j$)')
                ax_raw_scores_minus_quality_scores_and_observer_bias.set_ylabel(r'Test Subjects ($i$)')
                plt.colorbar(im, ax=ax_raw_scores_minus_quality_scores_and_observer_bias)
                if fig is not None:
                    fig.tight_layout()

    if do_plot == 'all' or 'quality_scores_vs_raw_scores' in do_plot:

        # TODO: visualize repetitions - currently taking mean over repetitions before plotting
        mtx = np.nanmean(dataset_reader.opinion_score_3darray, axis=2).T
        num_obs = mtx.shape[0]
        assert num_obs > 1, 'need snum_subj > 1 for subplots to work'

        min_lim = np.nanmin(mtx)
        max_lim = np.nanmax(mtx)

        nrows = int(math.floor(math.sqrt(num_obs)))
        ncols = int(math.ceil(num_obs / float(nrows)))

        for subjective_model, result in zip(subjective_models, results):
            if 'quality_scores' in result:
                fig, axs = plt.subplots(figsize=(ncols * 4, nrows * 4), ncols=ncols, nrows=nrows)
                quality_scores = result['quality_scores']
                label = subjective_model.TYPE
                for i_obs in range(num_obs):
                    assert num_obs > 1
                    ax = axs.flatten()[i_obs]
                    ax.set_title(f"#{i_obs + 1}")
                    raw_scores = mtx[i_obs, :]
                    ax.scatter(raw_scores, quality_scores, label=label)
                    ax.set_xlim([min_lim, max_lim])
                    ax.set_ylim([min_lim, max_lim])
                    ax.plot([min_lim, max_lim], [min_lim, max_lim], '-r')
                    ax.set_xlabel('Raw Score ($u_{ij}$)')
                    ax.set_ylabel('Recovered Quality Score ($\psi_j$)')
                    ax.legend()
                    ax.grid()
                fig.tight_layout()

    if do_plot == 'all' or 'quality_scores' in do_plot:
        # ===== plot quality scores =====
        bar_width = 0.4

        if 'ax_quality_scores' in ax_dict:
            ax_quality = ax_dict['ax_quality_scores']
            fig = None
        else:
            cols = None
            for result in results:
                if 'quality_scores' in result:
                    cols = len(result['quality_scores'])
                    break
            w, h = _get_plot_width_and_height(cols)
            fig, ax_quality = plt.subplots(figsize=(w, h), nrows=1)

        shift_count = 0
        for subjective_model, result in zip(subjective_models, results):
            if 'quality_scores' in result:
                quality = result['quality_scores']
                xs = range(len(quality))

                label = subjective_model.TYPE

                if 'bic' in result and 'reconstructions' in result:
                    label += ' [NBIC {:.2f}]'.format(result['bic'])

                if plot_type == 'bar':
                    ax_quality.bar(np.array(xs)+shift_count*bar_width, quality,
                                width=bar_width,
                                color=colors[shift_count],
                                label=label)

                elif plot_type == 'errorbar':
                    if 'quality_scores_ci95' in result:
                        try:
                            quality_error = result['quality_scores_ci95']
                            label += ' [avg CI {:.2f}]'.format(np.mean(np.array(quality_error[0]) +
                                                                       np.array(quality_error[1])))
                        except TypeError:
                            quality_error = None
                        ax_quality.errorbar(np.array(xs)+shift_count*bar_width+0.2, quality,
                                            yerr=quality_error, fmt='.', capsize=2,
                                            color=colors[shift_count],
                                            label=label)
                    else:
                        ax_quality.plot(np.array(xs)+shift_count*bar_width+0.2, quality, '.',
                                    color=colors[shift_count],
                                    label=label)
                else:
                    raise AssertionError("Unknown plot_type: {}".format(plot_type))

                ax_quality.set_xlabel(r'Video Stimuli ($j$)')
                ax_quality.set_title(r'Recovered Quality Score ($\psi_j$)')
                ax_quality.set_xlim([min(xs), max(xs)+1])
                shift_count += 1

        ax_quality.grid()
        ax_quality.legend(ncol=2, frameon=True)
        if fig is not None:
            fig.tight_layout()

    if do_plot == 'all' or 'subject_scores' in do_plot:

        # ===== plot subject bias and inconsistency =====
        bar_width = 0.4
        if 'ax_observer_bias' in ax_dict and 'ax_observer_inconsistency' in ax_dict:
            ax_bias = ax_dict['ax_observer_bias']
            ax_inconsty = ax_dict['ax_observer_inconsistency']
            fig = None
        else:
            cols = None
            for result in results:
                if 'observer_bias' in result:
                    cols = len(result['observer_bias'])
                    break
            w, h = _get_plot_width_and_height(cols)
            h = h * 2 + 2
            fig, (ax_bias, ax_inconsty) = plt.subplots(figsize=(w, h), nrows=2, ncols=1, sharex=True)

        if 'ax_rejected' in ax_dict:
            ax_rejected = ax_dict['ax_rejected']
        else:
            ax_rejected = None

        if 'ax_rejected_1st_stats' in ax_dict:
            ax_rejected_1st_stats = ax_dict['ax_rejected_1st_stats']
        else:
            ax_rejected_1st_stats = None

        if 'ax_rejected_2nd_stats' in ax_dict:
            ax_rejected_2nd_stats = ax_dict['ax_rejected_2nd_stats']
        else:
            ax_rejected_2nd_stats = None

        xs = None
        shift_count = 0
        my_xticks = None
        for subjective_model, result in zip(subjective_models, results):

            if 'observer_bias' in result:
                bias = result['observer_bias']
                xs = range(len(bias))

                if plot_type == 'bar':
                    ax_bias.bar(np.array(xs)+shift_count*bar_width, bias,
                                width=bar_width,
                                color=colors[shift_count],
                                label=subjective_model.TYPE)
                elif plot_type == 'errorbar':
                    if 'observer_bias_ci95' in result:
                        try:
                            bias_error = result['observer_bias_ci95']
                            label = '{} [avg CI {:.2f}]'.format(
                                subjective_model.TYPE, np.mean(np.array(bias_error[0]) +
                                                               np.array(bias_error[1])))
                        except TypeError:
                            bias_error = None
                            label = subjective_model.TYPE
                        ax_bias.errorbar(np.array(xs)+shift_count*bar_width+0.2, bias,
                                         yerr=bias_error, fmt='.', capsize=2,
                                         color=colors[shift_count],
                                         label=label)
                    else:
                        ax_bias.plot(np.array(xs)+shift_count*bar_width+0.2, bias, '.',
                                     color=colors[shift_count],
                                     label=subjective_model.TYPE)
                else:
                    raise AssertionError("Unknown plot_type: {}".format(plot_type))

                ax_inconsty.set_xlim([min(xs), max(xs)+1])
                ax_bias.set_title(r'Subject Bias ($\Delta_i$)')
                ax_bias.legend(ncol=1, frameon=True)

                if 'observers' in result:
                    observers = result['observers']
                    assert len(bias) == len(observers)
                    my_xticks = observers
                    plt.sca(ax_inconsty)
                    plt.xticks(np.array(xs) + 0.01, my_xticks, rotation=90)

            if 'observer_inconsistency' in result:
                inconsty = result['observer_inconsistency']

                xs = range(len(inconsty))

                if plot_type == 'bar':
                    ax_inconsty.bar(np.array(xs)+shift_count*bar_width, inconsty,
                                    width=bar_width,
                                    color=colors[shift_count],
                                    label=subjective_model.TYPE)
                elif plot_type == 'errorbar':
                    if 'observer_inconsistency_ci95' in result:
                        try:
                            inconsistency_error = result['observer_inconsistency_ci95']
                            label = '{} [avg CI {:.2f}]'.format(
                                subjective_model.TYPE, np.mean(np.array(inconsistency_error[0]) +
                                                               np.array(inconsistency_error[1])))
                        except TypeError:
                            inconsistency_error = None
                            label = subjective_model.TYPE
                        ax_inconsty.errorbar(np.array(xs)+shift_count*bar_width+0.2, inconsty,
                                             yerr=inconsistency_error, fmt='.', capsize=2,
                                             color=colors[shift_count],
                                             label=label)
                    else:
                        ax_inconsty.plot(np.array(xs)+shift_count*bar_width+0.2, inconsty, '.',
                                         color=colors[shift_count],
                                         label=subjective_model.TYPE)
                else:
                    raise AssertionError("Unknown plot_type: {}".format(plot_type))

                ax_inconsty.set_xlim([min(xs), max(xs)+1])
                ax_inconsty.set_title(r'Subject Inconsistency ($\upsilon_i$)')
                ax_inconsty.legend(ncol=1, frameon=True)

            if 'observer_rejected' in result and ax_rejected is not None:

                rejected = np.array(result['observer_rejected']).astype(int)

                xs = range(len(rejected))
                ax_rejected.bar(np.array(xs) + shift_count * bar_width, rejected,
                                width=bar_width,
                                color=colors[shift_count],
                                label=subjective_model.TYPE)
                ax_rejected.set_xlim([min(xs), max(xs)+1])
                ax_rejected.set_title(r'Subject Rejected')
                ax_rejected.legend(ncol=1, frameon=True)

            if 'observer_rejected_1st_stats' in result and ax_rejected_1st_stats is not None:

                rejected = result['observer_rejected_1st_stats']

                xs = range(len(rejected))
                ax_rejected_1st_stats.bar(np.array(xs) + shift_count * bar_width, rejected,
                                width=bar_width,
                                color=colors[shift_count],
                                label=subjective_model.TYPE)
                ax_rejected_1st_stats.set_xlim([min(xs), max(xs)+1])
                ax_rejected_1st_stats.set_title(r'Subject Rejected (1st stats)')
                ax_rejected_1st_stats.legend(ncol=1, frameon=True)

            if 'observer_rejected_2nd_stats' in result and ax_rejected_2nd_stats is not None:

                rejected = result['observer_rejected_2nd_stats']

                xs = range(len(rejected))
                ax_rejected_2nd_stats.bar(np.array(xs) + shift_count * bar_width, rejected,
                                width=bar_width,
                                color=colors[shift_count],
                                label=subjective_model.TYPE)
                ax_rejected_2nd_stats.set_xlim([min(xs), max(xs)+1])
                ax_rejected_2nd_stats.set_title(r'Subject Rejected (2nd stats)')
                ax_rejected_2nd_stats.legend(ncol=1, frameon=True)

            if 'observer_bias' in result or 'observer_inconsistency' in result or 'observer_rejected' in result:
                shift_count += 1

        if xs and my_xticks is None:
            my_xticks = list(map(lambda x: "#{}".format(x+1), xs))
            plt.sca(ax_inconsty)
            plt.xticks(np.array(xs) + 0.3, my_xticks, rotation=90)
            if ax_rejected is not None:
                plt.sca(ax_rejected)
                plt.xticks(np.array(xs) + 0.3, my_xticks, rotation=90)
            if ax_rejected_1st_stats is not None:
                plt.sca(ax_rejected_1st_stats)
                plt.xticks(np.array(xs) + 0.3, my_xticks, rotation=90)
            if ax_rejected_2nd_stats is not None:
                plt.sca(ax_rejected_2nd_stats)
                plt.xticks(np.array(xs) + 0.3, my_xticks, rotation=90)

        ax_bias.grid()
        ax_inconsty.grid()
        if ax_rejected is not None:
            ax_rejected.grid()
        if ax_rejected_1st_stats is not None:
            ax_rejected_1st_stats.grid()
        if ax_rejected_2nd_stats is not None:
            ax_rejected_2nd_stats.grid()
        if fig is not None:
            fig.tight_layout()

    if do_plot == 'all' or 'content_scores' in do_plot:

        # ===== plot content ambiguity =====
        bar_width = 0.4

        if 'ax_content_ambiguity' in ax_dict:
            ax_ambgty = ax_dict['ax_content_ambiguity']
            fig = None
        else:
            cols = None
            for result in results:
                if 'content_ambiguity' in result:
                    cols = len(result['content_ambiguity'])
                    break
            w, h = _get_plot_width_and_height(cols)
            fig, ax_ambgty = plt.subplots(figsize=(w, h), nrows=1)
        xs = None
        shift_count = 0
        for subjective_model, result in zip(subjective_models, results):
            if 'content_ambiguity' in result:
                ambgty = result['content_ambiguity']
                xs = range(len(ambgty))

                if plot_type == 'bar':
                    ax_ambgty.bar(np.array(xs)+shift_count*bar_width, ambgty,
                                  width=bar_width,
                                  color=colors[shift_count],
                                  label=subjective_model.TYPE)
                elif plot_type == 'errorbar':
                    if 'content_ambiguity_ci95' in result:
                        try:
                            ambiguity_error = result['content_ambiguity_ci95']
                            label = '{} [avg CI {:.2f}]'.format(
                                subjective_model.TYPE, np.mean(np.array(ambiguity_error[0]) +
                                                               np.array(ambiguity_error[1])))
                        except TypeError:
                            ambiguity_error = None
                            label = subjective_model.TYPE
                        ax_ambgty.errorbar(np.array(xs)+shift_count*bar_width+0.2, ambgty,
                                           yerr=ambiguity_error, fmt='.', capsize=2,
                                           color=colors[shift_count],
                                           label=label)
                    else:
                        ax_ambgty.plot(np.array(xs)+shift_count*bar_width+0.2, ambgty, '.',
                                       color=colors[shift_count],
                                       label=subjective_model.TYPE)
                else:
                    raise AssertionError("Unknown plot_type: {}".format(plot_type))

                shift_count += 1
                ax_ambgty.set_title(r'Content Ambiguity ($\rho_k$)')
                ax_ambgty.grid()
        if xs:
            my_xticks = ['' for _ in range(len(xs))]
            for ref_video in dataset_reader.dataset.ref_videos:
                my_xticks[ref_video['content_id']] = ref_video['content_name']
            rotation = 90
            plt.sca(ax_ambgty)
            plt.xticks(np.array(xs) + 0.01, my_xticks, rotation=rotation)
        ax_ambgty.legend(ncol=2, frameon=True)
        if fig is not None:
            fig.tight_layout()

    return dataset, subjective_models, results


def _get_imshow_width_and_height(rows, cols):
    w, h = cols / 30, rows / 30
    if w > 100:
        x = w // 100
        w, h = w / x, h / x
    if h > 100:
        x = h // 100
        w, h = w / x, h / x
    w, h = w + 2, h + 2
    return w, h


def _get_plot_width_and_height(num: Union[int, None]) -> [int, int]:
    if num is None:
        return 10, 3
    else:
        return max(np.sqrt(num) + 2, 10), 3


def format_output_of_run_subjective_models(dataset, subjective_models, results):

    assert len(subjective_models) == len(results)
    for result in results:
        assert 'quality_scores' in result
        assert 'dis_video_names' in result
        assert len(result['dis_video_names']) == len(result['quality_scores'])
        if 'quality_scores_std' in result:
            assert len(result['quality_scores']) == len(result['quality_scores_std'])
        if 'quality_scores_ci95' in result:
            assert len(result['quality_scores_ci95']) == 2
            assert len(result['quality_scores']) == len(list(zip(*result['quality_scores_ci95'])))

        if 'observer_bias' in result:
            if 'observers' in result:
                assert len(result['observers']) == len(result['observer_bias'])
            else:
                result['observers'] = [f'observer{o}' for o in range(len(result['observer_bias']))]
            if 'observer_bias_std' in result:
                assert len(result['observer_bias']) == len(result['observer_bias_std'])
            if 'observer_bias_ci95' in result:
                assert len(result['observer_bias_ci95']) == 2
                assert len(result['observer_bias']) == len(list(zip(*result['observer_bias_ci95'])))
            if 'observer_inconsistency' in result:
                assert len(result['observer_bias']) == len(result['observer_inconsistency'])
            if 'observer_inconsistency_std' in result:
                assert len(result['observer_bias']) == len(result['observer_inconsistency_std'])
            if 'observer_inconsistency_ci95' in result:
                assert len(result['observer_inconsistency_ci95']) == 2
                assert len(result['observer_bias']) == len(list(zip(*result['observer_inconsistency_ci95'])))
            if 'observer_scores_mean' in result:
                assert len(result['observer_bias']) == len(result['observer_scores_mean'])
            if 'observer_scores_std' in result:
                assert len(result['observer_bias']) == len(result['observer_scores_std'])

        if 'content_ambiguity' in result:
            dict_contentid_content = dict()
            for ref_video in dataset.ref_videos:
                dict_contentid_content[ref_video['content_id']] = \
                    f"{ref_video['content_name']}"
            result['contents'] = [dict_contentid_content[k] for k in sorted(dict_contentid_content.keys())]
            assert len(result['contents']) == len(result['content_ambiguity'])
            if 'content_ambiguity_std' in result:
                assert len(result['content_ambiguity']) == len(result['content_ambiguity_std'])
            if 'content_ambiguity_ci95' in result:
                assert len(result['content_ambiguity_ci95']) == 2
                assert len(result['content_ambiguity']) == len(list(zip(*result['content_ambiguity_ci95'])))

    output = dict()
    for subjective_model, result in zip(subjective_models, results):
        for idx, dis_video_name in enumerate(result['dis_video_names']):
            if 'dis_video_name' in output.setdefault('dis_videos', dict()).setdefault(idx, dict()):
                assert output.setdefault('dis_videos', dict()).setdefault(idx, dict())['dis_video_name'] == dis_video_name
            else:
                output.setdefault('dis_videos', dict()).setdefault(idx, dict())['dis_video_name'] = dis_video_name
        for idx, quality_score in enumerate(result['quality_scores']):
            output.setdefault('dis_videos', dict()).setdefault(idx, dict()).setdefault('models', dict()).setdefault(subjective_model.TYPE, dict())['quality_score'] = quality_score
        if 'quality_scores_std' in result:
            for idx, quality_score_std in enumerate(result['quality_scores_std']):
                output.setdefault('dis_videos', dict()).setdefault(idx, dict()).setdefault('models', dict()).setdefault(subjective_model.TYPE, dict())['quality_score_std'] = quality_score_std
        if 'quality_scores_ci95' in result:
            for idx, quality_score_ci95 in enumerate(list(zip(*result['quality_scores_ci95']))):
                output.setdefault('dis_videos', dict()).setdefault(idx, dict()).setdefault('models', dict()).setdefault(subjective_model.TYPE, dict())['quality_score_ci95'] = quality_score_ci95

        if 'observer_bias' in result:
            if 'observers' in result:
                for idx, observer in enumerate(result['observers']):
                    if 'observer' in output.setdefault('observers', dict()).setdefault(idx, dict()):
                        assert output.setdefault('observers', dict()).setdefault(idx, dict())['observer'] == observer
                    else:
                        output.setdefault('observers', dict()).setdefault(idx, dict())['observer'] = observer
            for idx, observer_bias in enumerate(result['observer_bias']):
                output.setdefault('observers', dict()).setdefault(idx, dict()).setdefault('models', dict()).setdefault(subjective_model.TYPE, dict())['observer_bias'] = observer_bias
            if 'observer_bias_std' in result:
                for idx, observer_bias_std in enumerate(result['observer_bias_std']):
                    output.setdefault('observers', dict()).setdefault(idx, dict()).setdefault('models', dict()).setdefault(subjective_model.TYPE, dict())['observer_bias_std'] = observer_bias_std
            if 'observer_bias_ci95' in result:
                for idx, observer_bias_ci95 in enumerate(list(zip(*result['observer_bias_ci95']))):
                    output.setdefault('observers', dict()).setdefault(idx, dict()).setdefault('models', dict()).setdefault(subjective_model.TYPE, dict())['observer_bias_ci95'] = observer_bias_ci95
            if 'observer_inconsistency' in result:
                for idx, observer_inconsistency in enumerate(result['observer_inconsistency']):
                    output.setdefault('observers', dict()).setdefault(idx, dict()).setdefault('models', dict()).setdefault(subjective_model.TYPE, dict())['observer_inconsistency'] = observer_inconsistency
            if 'observer_inconsistency_std' in result:
                for idx, observer_inconsistency_std in enumerate(result['observer_inconsistency_std']):
                    output.setdefault('observers', dict()).setdefault(idx, dict()).setdefault('models', dict()).setdefault(subjective_model.TYPE, dict())['observer_inconsistency_std'] = observer_inconsistency_std
            if 'observer_inconsistency_ci95' in result:
                for idx, observer_inconsistency_ci95 in enumerate(list(zip(*result['observer_inconsistency_ci95']))):
                    output.setdefault('observers', dict()).setdefault(idx, dict()).setdefault('models', dict()).setdefault(subjective_model.TYPE, dict())['observer_inconsistency_ci95'] = observer_inconsistency_ci95
            if 'observer_scores_mean' in result:
                for idx, observer_scores_mean in enumerate(result['observer_scores_mean']):
                    output.setdefault('observers', dict()).setdefault(idx, dict()).setdefault('models', dict()).setdefault(subjective_model.TYPE, dict())['observer_scores_mean'] = observer_scores_mean
            if 'observer_scores_std' in result:
                for idx, observer_scores_std in enumerate(result['observer_scores_std']):
                    output.setdefault('observers', dict()).setdefault(idx, dict()).setdefault('models', dict()).setdefault(subjective_model.TYPE, dict())['observer_scores_std'] = observer_scores_std

        if 'content_ambiguity' in result:
            if 'contents' in result:
                for idx, content in enumerate(result['contents']):
                    if 'content' in output.setdefault('contents', dict()).setdefault(idx, dict()):
                        assert output.setdefault('contents', dict()).setdefault(idx, dict())['content'] == content
                    else:
                        output.setdefault('contents', dict()).setdefault(idx, dict())['content'] = content
            for idx, content_ambiguity in enumerate(result['content_ambiguity']):
                output.setdefault('contents', dict()).setdefault(idx, dict()).setdefault('models', dict()).setdefault(subjective_model.TYPE, dict())['content_ambiguity'] = content_ambiguity
            if 'content_ambiguity_std' in result:
                for idx, content_ambiguity_std in enumerate(result['content_ambiguity_std']):
                    output.setdefault('contents', dict()).setdefault(idx, dict()).setdefault('models', dict()).setdefault(subjective_model.TYPE, dict())['content_ambiguity_std'] = content_ambiguity_std
            if 'content_ambiguity_ci95' in result:
                for idx, content_ambiguity_ci95 in enumerate(list(zip(*result['content_ambiguity_ci95']))):
                    output.setdefault('contents', dict()).setdefault(idx, dict()).setdefault('models', dict()).setdefault(subjective_model.TYPE, dict())['content_ambiguity_ci95'] = content_ambiguity_ci95

        if 'aic' in result:
            output.setdefault('stats', dict()).setdefault('models', dict()).setdefault(subjective_model.TYPE, dict())['aic'] = result['aic']
        if 'bic' in result:
            output.setdefault('stats', dict()).setdefault('models', dict()).setdefault(subjective_model.TYPE, dict())['bic'] = result['bic']

    if 'dis_videos' in output:
        output['dis_videos'] = list(output['dis_videos'].values())
    if 'observers' in output:
        output['observers'] = list(output['observers'].values())
    if 'contents' in output:
        output['contents'] = list(output['contents'].values())

    return output


def visualize_pc_dataset(dataset_filepath):

    dataset = import_python_file(dataset_filepath)
    dataset_reader = PairedCompDatasetReader(dataset)
    tensor_pvs_pvs_subject = dataset_reader.opinion_score_3darray

    plt.figure()
    mtx_pvs_pvs = np.nansum(tensor_pvs_pvs_subject, axis=2) \
                  / (np.nansum(tensor_pvs_pvs_subject, axis=2) +
                     np.nansum(tensor_pvs_pvs_subject, axis=2).transpose())
    plt.imshow(mtx_pvs_pvs, interpolation='nearest')
    plt.title(r'Paired Comparison Winning Rate')
    plt.ylabel(r"PVS ($j$)")
    plt.xlabel(r"PVS ($j'$) [Compared Against]")
    plt.set_cmap('jet')
    plt.colorbar()
    plt.tight_layout()


def validate_with_synthetic_dataset(synthetic_dataset_reader_class,
                                    subjective_model_classes,
                                    dataset_filepath,
                                    synthetic_result,
                                    ax_dict,
                                    **more):
        ret = {}

        color_dict = more['color_dict'] if 'color_dict' in more else {}

        marker_dict = more['marker_dict'] if 'marker_dict' in more else {}

        output_synthetic_dataset_filepath = more['output_synthetic_dataset_filepath'] \
            if 'output_synthetic_dataset_filepath' in more else None

        missing_probability = more['missing_probability'] if 'missing_probability' in more else None
        assert missing_probability is None or 0 <= missing_probability < 1

        measure_runtime = more['measure_runtime'] if 'measure_runtime' in more else False
        assert isinstance(measure_runtime, bool)

        dataset = import_python_file(dataset_filepath)
        dataset_reader = synthetic_dataset_reader_class(dataset, input_dict=synthetic_result)

        if missing_probability is not None:
            dataset = dataset_reader.to_dataset()
            synthetic_result2 = copy.deepcopy(synthetic_result)
            synthetic_result2.update({'missing_probability': missing_probability})
            dataset_reader = MissingDataRawDatasetReader(dataset, input_dict=synthetic_result2)

        if output_synthetic_dataset_filepath is not None:
            dataset_reader.write_out_dataset(dataset_reader.to_dataset(), output_synthetic_dataset_filepath)
            ret['output_synthetic_dataset_filepath'] = output_synthetic_dataset_filepath

        subjective_models = map(
            lambda subjective_model_class: subjective_model_class(dataset_reader),
            subjective_model_classes
        )

        def run_modeling(subjective_model):
            if measure_runtime:
                with Timer() as t:
                    ret = subjective_model.run_modeling(**more)
                ret['runtime'] = t.interval
            else:
                ret = subjective_model.run_modeling(**more)
            return ret

        results = list(map(
            lambda subjective_model: run_modeling(subjective_model),
            subjective_models
        ))

        if ax_dict is None:
            ax_dict = dict()

        do_errorbar = more['do_errorbar'] if 'do_errorbar' in more else False
        assert isinstance(do_errorbar, bool)

        ret['results'] = dict()
        for subjective_model_class, result in zip(subjective_model_classes, results):
            ret['results'][subjective_model_class.TYPE] = result

        for ax in ax_dict.values():
            ax.set_xlabel('Synthetic')
            ax.set_ylabel('Recovered')
            ax.grid()

        for subjective_model_class, result, idx in zip(subjective_model_classes, results, range(len(results))):

            model_name = subjective_model_class.TYPE

            if 'quality_scores' in result and 'quality_scores' in synthetic_result and 'quality_scores_ci95' in result:
                ci_perc = get_ci_percentage(synthetic_result, result, 'quality_scores', 'quality_scores_ci95')
                ret['results'][model_name]['quality_scores_ci_perc'] = ci_perc

            if 'observer_bias' in result and 'observer_bias' in synthetic_result and 'observer_bias_ci95' in result:
                ci_perc = get_ci_percentage(synthetic_result, result, 'observer_bias', 'observer_bias_ci95')
                ret['results'][model_name]['observer_bias_ci_perc'] = ci_perc

            if 'observer_inconsistency' in result and 'observer_inconsistency' in synthetic_result \
                    and 'observer_inconsistency_ci95' in result:
                ci_perc = get_ci_percentage(synthetic_result, result,
                                            'observer_inconsistency', 'observer_inconsistency_ci95')
                ret['results'][model_name]['observer_inconsistency_ci_perc'] = ci_perc

            if 'quality_scores' in ax_dict:
                ax = ax_dict['quality_scores']
                ax.set_title(r'Quality Score ($\psi_j$)')
                if 'quality_scores' in result and 'quality_scores' in synthetic_result:
                    color = color_dict[model_name] if model_name in color_dict else 'black'
                    marker = marker_dict[model_name] if model_name in marker_dict else '.'
                    x = synthetic_result['quality_scores']
                    y = result['quality_scores']
                    if 'quality_scores_ci95' in result:
                        yerr = result['quality_scores_ci95']
                        ci_perc = get_ci_percentage(synthetic_result, result,  'quality_scores', 'quality_scores_ci95')
                    else:
                        yerr = None
                        ci_perc=None
                    if do_errorbar is True and 'quality_scores_ci95' in result:
                        ax.errorbar(x, y, yerr=yerr, color=color, capsize=2, marker=marker, linestyle='None',
                                    label='{sm} (RMSE {rmse:.4f}, CI% {ci_perc:.1f})'.format(
                                        sm=model_name,
                                        rmse=RmsePerfMetric(x, y).evaluate(enable_mapping=False)['score'],
                                        ci_perc=ci_perc,
                                    ))
                    else:
                        ax.scatter(x, y, color=color, marker=marker,
                                   label='{sm} (RMSE {rmse:.4f})'.format(
                                       sm=model_name, rmse=RmsePerfMetric(x, y).evaluate(enable_mapping=False)['score']))
                    ax.legend()
                    diag_line = np.arange(min(x), max(x), step=0.01)
                    ax.plot(diag_line, diag_line, '-', color='gray', )

            if 'quality_scores_std' in ax_dict:
                ax = ax_dict['quality_scores_std']
                ax.set_title(r'Std of Quality Score ($\sigma(\psi_j)$)')
                if 'quality_scores_std' in result and 'quality_scores_std' in synthetic_result:
                    color = color_dict[model_name] if model_name in color_dict else 'black'
                    marker = marker_dict[model_name] if model_name in marker_dict else '.'
                    x = synthetic_result['quality_scores_std']
                    y = result['quality_scores_std']
                    ax.scatter(x, y, color=color, marker=marker,
                               label='{sm} (RMSE {rmse:.4f})'.format(
                                   sm=model_name, rmse=RmsePerfMetric(x, y).evaluate(enable_mapping=False)['score']))
                    ax.legend()
                    diag_line = np.arange(min(x), max(x), step=0.01)
                    ax.plot(diag_line, diag_line, '-', color='gray', )

            if 'observer_bias' in ax_dict:
                ax = ax_dict['observer_bias']
                ax.set_title(r'Subject Bias ($\Delta_i$)')
                if 'observer_bias' in result and 'observer_bias' in synthetic_result:
                    color = color_dict[model_name] if model_name in color_dict else 'black'
                    marker = marker_dict[model_name] if model_name in marker_dict else '.'
                    x = synthetic_result['observer_bias']
                    y = result['observer_bias']
                    if 'observer_bias_ci95' in result:
                        yerr = result['observer_bias_ci95']
                        ci_perc = get_ci_percentage(synthetic_result, result, 'observer_bias', 'observer_bias_ci95')
                    else:
                        yerr = None
                        ci_perc = None
                    min_xy = np.min([len(x),len(y)])
                    x = x[:min_xy]
                    y = y[:min_xy]
                    if do_errorbar is True and 'observer_bias_ci95' in result:
                        ax.errorbar(x, y, yerr=yerr, color=color, capsize=2, marker=marker, linestyle='None',
                                    label='{sm} (RMSE {rmse:.4f}, CI% {ci_perc:.1f})'.format(
                                        sm=model_name,
                                        rmse=RmsePerfMetric(x, y).evaluate(enable_mapping=False)['score'],
                                        ci_perc=ci_perc,
                                    ))
                    else:
                        ax.scatter(x, y, color=color, marker=marker,
                                    label='{sm} (RMSE {rmse:.4f})'.format(
                                        sm=model_name,
                                        rmse=RmsePerfMetric(x, y).evaluate(enable_mapping=False)['score']))
                    ax.legend()
                    diag_line = np.arange(min(x), max(x), step=0.01)
                    ax.plot(diag_line, diag_line, '-', color='gray', )

            if 'observer_inconsistency' in ax_dict:
                ax = ax_dict['observer_inconsistency']
                ax.set_title(r'Subject Inconsistency ($\upsilon_i$)')
                if 'observer_inconsistency' in result and 'observer_inconsistency' in synthetic_result:
                    color = color_dict[model_name] if model_name in color_dict else 'black'
                    marker = marker_dict[model_name] if model_name in marker_dict else '.'
                    x = synthetic_result['observer_inconsistency']
                    y = result['observer_inconsistency']
                    if 'observer_inconsistency_ci95' in result:
                        yerr = np.array(result['observer_inconsistency_ci95'])
                        ci_perc = get_ci_percentage(synthetic_result, result,
                                                    'observer_inconsistency', 'observer_inconsistency_ci95')
                    else:
                        yerr = None
                        ci_perc = None
                    min_xy = np.min([len(x),len(y)])
                    x = x[:min_xy]
                    y = y[:min_xy]
                    if do_errorbar is True and 'observer_inconsistency_ci95' in result:
                        ax.errorbar(x, y, yerr=yerr, color=color, capsize=2, marker=marker, linestyle='None',
                                    label='{sm} (RMSE {rmse:.4f}, CI% {ci_perc:.1f})'.format(
                                        sm=model_name,
                                        rmse=RmsePerfMetric(x, y).evaluate(enable_mapping=False)['score'],
                                        ci_perc=ci_perc,
                                    ))
                    else:
                        ax.scatter(x, y, color=color, marker=marker,
                                   label='{sm} (RMSE {rmse:.4f})'.format(
                                       sm=model_name, rmse=RmsePerfMetric(x, y).evaluate(enable_mapping=False)['score']))
                    ax.legend()
                    diag_line = np.arange(min(x), max(x), step=0.01)
                    ax.plot(diag_line, diag_line, '-', color='gray', )

            if 'quality_ambiguity' in ax_dict:
                ax = ax_dict['quality_ambiguity']
                ax.set_title(r'Quality Ambiguity ($\phi_j$)')
                if 'quality_ambiguity' in result and 'quality_ambiguity' in synthetic_result:
                    color = color_dict[model_name] if model_name in color_dict else 'black'
                    marker = marker_dict[model_name] if model_name in marker_dict else '.'
                    x = synthetic_result['quality_ambiguity']
                    y = result['quality_ambiguity']
                    min_xy = np.min([len(x),len(y)])
                    x = x[:min_xy]
                    y = y[:min_xy]
                    ax.scatter(x, y, color=color, marker=marker,
                               label='{sm} (RMSE {rmse:.4f})'.format(
                                   sm=model_name, rmse=RmsePerfMetric(x, y).evaluate(enable_mapping=False)['score']))
                    ax.legend()
                    diag_line = np.arange(min(x), max(x), step=0.01)
                    ax.plot(diag_line, diag_line, '-', color='gray', )

        return ret


def get_ci_percentage(synthetic_result, result, key, errkey):
    xs = synthetic_result[key]
    ys = result[key]
    ys_ci95 = np.array(result[errkey])
    ys_ci95_shape = ys_ci95.shape
    assert ys_ci95_shape[0] == 2
    assert len(xs) == len(ys) == ys_ci95_shape[1]
    ind_in_ci = [y - y_ci95_l <= x <= y + y_ci95_u for x, y, y_ci95_l, y_ci95_u in zip(xs, ys, ys_ci95[0], ys_ci95[1])]
    ci_perc = sum(ind_in_ci) / len(ind_in_ci) * 100
    return ci_perc


def get_sample_stats(datasets, subjective_model_classes, do_plot=False, plot_type='bar', subj_fraction=None, random_seed=None):

    resultss = []  # dataset x subjective_model
    for dataset in datasets:
        if do_plot:
            fig_raw_scores, ax_raw_scores = plt.subplots(figsize=(7, 3))
            fig_quality_scores, ax_quality_scores = plt.subplots(figsize=(12, 3.5), nrows=1)

            # fig_bias_inconsty, [ax_bias, ax_inconsty] = plt.subplots(figsize=(6, 5), nrows=2, ncols=1, sharex=True)
            # fig_rejected, ax_rejected = plt.subplots(figsize=(6, 3), nrows=1)
            fig_bias_inconsty, [ax_bias, ax_inconsty, ax_rejected] = plt.subplots(figsize=(6, 7), nrows=3, ncols=1, sharex=True)
            fig_rejected = None

            ax_dict = {
                'ax_raw_scores': ax_raw_scores,
                'ax_quality_scores': ax_quality_scores,
                'ax_observer_bias': ax_bias,
                'ax_observer_inconsistency': ax_inconsty,
                'ax_rejected': ax_rejected,
            }
            do_plot_list = [
                'raw_scores',
                'quality_scores',
                'subject_scores'
            ]
        else:
            fig_raw_scores = None
            fig_quality_scores = None
            fig_bias_inconsty = None
            fig_rejected = None
            ax_dict = {}
            do_plot_list = []

        dataset_filepath = dataset['path']
        dataset_reader_info_dict = {}
        if subj_fraction is None:
            dataset, subjective_models, results = run_subjective_models(
                dataset_filepath=dataset_filepath,
                subjective_model_classes=subjective_model_classes,
                normalize_final=False,  # True or False
                do_plot=do_plot_list,
                plot_type=plot_type,
                ax_dict=ax_dict,
                dataset_reader_info_dict=dataset_reader_info_dict
            )
            resultss.append(results)
        else:
            assert 0.0 < subj_fraction <= 1.0

            if dataset_filepath.endswith('.py'):
                dataset = import_python_file(dataset_filepath)
            elif dataset_filepath.endswith('.json'):
                dataset = import_json_file(dataset_filepath)
            else:
                raise AssertionError("Unknown input type, must be .py or .json")
            dataset_reader = RawDatasetReader(dataset, input_dict=dataset_reader_info_dict)
            num_subj = dataset_reader.num_observers
            frac_num_subj = int(num_subj * subj_fraction)
            if random_seed is not None:
                    np.random.seed(random_seed)
            selected_subjects = np.random.choice(list(range(num_subj)), size=frac_num_subj, replace=False)
            dataset_reader_info_dict_new = dataset_reader_info_dict.copy()
            dataset_reader_info_dict_new['selected_subjects'] = selected_subjects
            dataset, subjective_models, results = run_subjective_models(
                dataset_filepath=dataset_filepath,
                subjective_model_classes=subjective_model_classes,
                normalize_final=False,  # True or False
                do_plot=do_plot_list,
                plot_type=plot_type,
                ax_dict=ax_dict,
                dataset_reader_info_dict=dataset_reader_info_dict_new,
                dataset_reader_class=SelectSubjectRawDatasetReader,
            )
            resultss.append(results)

        if fig_raw_scores is not None:
            fig_raw_scores.tight_layout()
        if fig_quality_scores is not None:
            fig_quality_scores.tight_layout()
        if fig_bias_inconsty is not None:
            fig_bias_inconsty.tight_layout()
        if fig_rejected is not None:
            fig_rejected.tight_layout()

    return resultss


def plot_scatter_target_vs_compared_models(target_models, compared_models, datasets,
                                           target_subj_fraction=None,
                                           compared_subj_fraction=None,
                                           random_seed=None,
                                           get_sample_stats_wrapper_method=None
                                           ):

    if get_sample_stats_wrapper_method is None:
        get_sample_stats_wrapper_method = _get_sample_stats_wrapper

    for dataset in datasets:

        target_resultss = []
        for model in target_models:
            results_target = get_sample_stats_wrapper_method(
                dataset,
                target_subj_fraction,
                model, random_seed)
            target_resultss.append(results_target)

        compared_resultss = []
        for model in compared_models:
            results_compared = get_sample_stats_wrapper_method(
                dataset,
                compared_subj_fraction,
                model, random_seed)
            compared_resultss.append(results_compared)

        fig, axss = plt.subplots(ncols=len(compared_resultss), nrows=2,
                                 figsize=[5 * len(compared_resultss), 7],
                                 gridspec_kw={'height_ratios': [3, 1]})
        axss = axss.T
        if len(compared_resultss) == 1:
            axss = np.array([axss])

        for target_model, target_results, compapred_model, compared_results, axs in zip(target_models, target_resultss, compared_models, compared_resultss, axss):
            target_results = target_results[0]
            compared_results = compared_results[0]
            for target_result, compared_result \
                    in zip(target_results, compared_results):
                ax_scatter, ax_hist = axs
                xs = target_result['quality_scores']
                ys = compared_result['quality_scores']
                diffs = np.array(ys) - np.array(xs)
                mean_diff = np.mean(diffs)
                std_diff = np.std(diffs)
                plcc = scipy.stats.pearsonr(xs, ys)[0]
                srocc = scipy.stats.spearmanr(xs, ys)[0]
                label = os.path.splitext(os.path.basename(dataset['path']))[0][:30]
                compared_tag = compapred_model
                if compared_subj_fraction is not None:
                    compared_tag += f" {int(compared_subj_fraction * 100)}% data"
                target_tag = target_model
                if target_subj_fraction is not None:
                    target_tag += f" {int(target_subj_fraction * 100)}% data"
                ax_scatter.scatter(xs, ys, alpha=0.2, label=label)
                ax_scatter.set_xlabel(target_tag)
                ax_scatter.set_ylabel(compared_tag)
                ax_scatter.grid()
                ax_scatter.legend()
                ax_hist.hist(diffs, label=label)
                ax_hist.set_title(
                    f"plcc {plcc:.3f}, srocc {srocc:.3f}, diff: mean {mean_diff:.3f}, std {std_diff:.3f}")
                ax_hist.set_xlabel(
                    f"diff: {compared_tag} vs. {target_tag}")
                ax_hist.set_ylabel(f"No. ocurrences")
                ax_hist.grid()
                ax_hist.legend()
            plt.tight_layout()


def _get_sample_stats_wrapper(dataset, subj_fraction, model, random_seed):
    model_class = SubjectiveModel.find_subclass(model)
    results = get_sample_stats(
        [dataset],
        [model_class],
        do_plot=False,
        subj_fraction=subj_fraction,
        random_seed=random_seed,
    )  # dataset x subjective_model
    for result in results:
        if 'raw_scores' in result[0]:
            del result[0]['raw_scores']
        if 'reconstructions' in result[0]:
            del result[0]['reconstructions']

    return results


@persist_to_dir(SurealConfig.workdir_path('_get_sample_stats_wrapper'))
def _get_sample_stats_wrapper_persistent(dataset, subj_fraction, model, random_seed):
    return _get_sample_stats_wrapper(dataset, subj_fraction, model, random_seed)



import os

__copyright__ = "Copyright 2016-2018, Netflix, Inc."
__license__ = "Apache, Version 2.0"

PYTHON_ROOT = os.path.join(os.path.dirname(os.path.realpath(__file__)), '..')


class SurealConfig(object):

    @classmethod
    def root_path(cls, *components):
        return os.path.join(PYTHON_ROOT, *components)

    @classmethod
    def workspace_path(cls, *components):
        return cls.root_path('workspace', *components)

    @classmethod
    def workdir_path(cls, *components):
        return cls.root_path('workspace', 'workdir', *components)

    @classmethod
    def resource_path(cls, *components):
        return cls.root_path('resource', *components)

    @classmethod
    def test_resource_path(cls, *components):
        return cls.root_path('test', 'resource', *components)


class DisplayConfig(object):

    @staticmethod
    def show(**kwargs):
        import matplotlib.pyplot as plt
        if 'write_to_dir' in kwargs:
            format = kwargs['format'] if 'format' in kwargs else 'png'
            filedir = kwargs['write_to_dir'] if kwargs['write_to_dir'] is not None else SurealConfig.workspace_path('output')
            os.makedirs(filedir, exist_ok=True)
            for fignum in plt.get_fignums():
                fig = plt.figure(fignum)
                fig.savefig(os.path.join(filedir, str(fignum) + '.' + format), format=format)
        else:
            plt.show()



import os
import json
import hashlib
import sys
import warnings
from functools import partial

__copyright__ = "Copyright 2016-2018, Netflix, Inc."
__license__ = "Apache, Version 2.0"


def persist_to_file(file_name):
    """
    Cache (or persist) returned value of function in a json file .
    """

    def decorator(original_func):

        if not os.path.exists(file_name):
            cache = {}
        else:
            try:
                cache = json.load(open(file_name, 'rt'))
            except (IOError, ValueError):
                sys.exit(1)

        def new_func(*args):
            h = hashlib.sha1(str(original_func.__name__) + str(args)).hexdigest()
            if h not in cache:
                cache[h] = original_func(*args)
                json.dump(cache, open(file_name, 'wt'))
            return cache[h]

        return new_func

    return decorator


def persist_to_dir(dir_name):
    """
    Cache (or persist) returned value of function in a directory of files.
    """

    def decorator(original_func):

        def new_func(*args):
            h = hashlib.sha1((str(original_func.__name__) + str(args)).encode('utf-8')).hexdigest()
            file_name = os.path.join(dir_name, h)
            if not os.path.exists(file_name):
                if not os.path.exists(dir_name):
                    os.makedirs(dir_name)
                res = original_func(*args)
                json.dump(res, open(file_name, 'wt'))
            else:
                res = json.load(open(file_name, 'rt'))
            return res

        return new_func

    return decorator


def deprecated(func):
    """
    Mark a function as deprecated.
    It will result in a warning being emmitted when the function is used.
    """

    def new_func(*args, **kwargs):
        warnings.simplefilter('always', DeprecationWarning) # turn off filter
        warnings.warn("Call to deprecated function {}.".format(func.__name__),
                      category=DeprecationWarning, stacklevel=2)
        warnings.simplefilter('default', DeprecationWarning) # reset filter
        return func(*args, **kwargs)

    new_func.__name__ = func.__name__
    new_func.__doc__ = func.__doc__
    new_func.__dict__.update(func.__dict__)
    return new_func


class memoized(object):
    """ Decorator. Caches a function's return value each time it is called.
    If called later with the same arguments, the cached value is returned
    (not reevaluated).

    memoized is similar to persist, but if applied to
    class methods, persist will cache on a per-class basis, while memoized
    will cache on a per-object basis.

    Taken from: https://wiki.python.org/moin/PythonDecoratorLibrary#Memoize
    """

    def __init__(self, func):
        self.func = func
        self.cache = {}

    def __call__(self, *args):
        if not hasattr(args, "__hash__"):
            return self.func(*args)
        if args in self.cache:
            return self.cache[args]
        else:
            value = self.func(*args)
            self.cache[args] = value
            return value

    def __repr__(self):
        """ Return the function's docstring. """
        return self.func.__doc__

    def __get__(self, obj, objtype):
        """ Support instance methods. """
        return partial(self.__call__, obj)


def override(interface_class):
    def overrider(method):
        assert(method.__name__ in dir(interface_class)), \
            f"{method.__name__} does not override any method " \
            f"in {interface_class.__name__}"
        return method
    return overrider



import numpy as np
import scipy
import scipy.signal
from .inverse import inversefunc
import warnings

# import multiprocessing
# pool = multiprocessing.Pool()

from .misc import parallel_map

__copyright__ = "Copyright 2016-2018, Netflix, Inc."
__license__ = "Apache, Version 2.0"


def vectorized_gaussian(xs, locs, scales):
    # f = lambda x, scale: stats.norm.pdf(x, scale=scale)
    # ff = np.vectorize(f)
    # return ff(xs - locs, scales)

    return 1.0 / np.sqrt(2 * np.pi) / scales * np.exp(- (xs - locs)**2 / (2* scales**2))


def vectorized_logistic(xs, locs, scales):
    # f = lambda x, scale: stats.logistic.pdf(x, scale=scale)
    # ff = np.vectorize(f)
    # return ff(xs - locs, scales)

    return 1.0 / 4.0 / scales / np.cosh((xs - locs) / 2.0 / scales)**2


def sech(x):
    return 2. / ( np.exp(x) + np.exp(-x) )


def vectorized_convolution_of_two_logistics(xs, locs1, scales1, locs2, scales2):

    f = lambda x, loc1, scale1, loc2, scale2: \
        ConvolveTwoPdf(
            lambda x: 1.0 / 4.0 / scale1 * sech(x / 2.0 / scale1)**2,
            lambda x: 1.0 / 4.0 / scale2 * sech(x / 2.0 / scale2)**2,

            # lambda x: 1.0 / 4.0 / scale1 / np.cosh(x / 2.0 / scale1)**2,
            # lambda x: 1.0 / 4.0 / scale2 / np.cosh(x / 2.0 / scale2)**2,

            # lambda x: 1.0 / np.sqrt(2 * np.pi * (scale1**2) * (np.pi**2 / 3.)) * np.exp(- x**2 / (2* (scale1**2) * (np.pi**2 / 3.))), # test gaussian
            # lambda x: 1.0 / np.sqrt(2 * np.pi * (scale2**2) * (np.pi**2 / 3.)) * np.exp(- x**2 / (2* (scale2**2) * (np.pi**2 / 3.))), # test gaussian

            f_truncation=1e-12,
            g_truncation=1e-12,
            delta=3.0e-3,
        ).pdf(x - loc1 - loc2)

    # # === way 1: parallel_map (each job too small, bottlenecked by passing context) ===
    # f2 = lambda x: f(*x)
    # xshape = xs.shape
    # assert xshape == locs1.shape == scales1.shape == locs2.shape == scales2.shape
    # res = parallel_map(f2, zip(xs.ravel(), locs1.ravel(), scales1.ravel(), locs2.ravel(), scales2.ravel()), pause_sec=None)
    # return np.reshape(res, xshape)

    # # === way 2: vectorize (sequential execution) ===
    # ff = np.vectorize(f)
    # return ff(xs, locs1, scales1, locs2, scales2)

    # === way 3: parallel map combined with vectorize (best speed) ===
    ff = np.vectorize(f)
    ff2 = lambda x: ff(*x)
    xshape = xs.shape
    assert xshape == locs1.shape == scales1.shape == locs2.shape == scales2.shape
    with np.errstate(over='ignore'):
        res = parallel_map(ff2, zip(xs, locs1, scales1, locs2, scales2), pause_sec=None)
    return np.array(res)

    # === test: test one gaussian ===
    # return 1.0 / np.sqrt(2 * np.pi * (scales1**2 + scales2**2) * (np.pi**2 / 3.)) * np.exp(- (xs - locs1 - locs2)**2 / (2* (scales1**2 + scales2**2) * (np.pi**2 / 3.)))


def convolution_of_two_uniforms(x, loc1, s1, loc2, s2):
    """
    >>> convolution_of_two_uniforms(-2, 0, 1, 0, 2)
    0.0
    >>> convolution_of_two_uniforms(-1.5, 0, 1, 0, 2)
    0.0
    >>> convolution_of_two_uniforms(-1.49, 0, 1, 0, 2)
    0.0050000000000000044
    >>> convolution_of_two_uniforms(-0.51, 0, 1, 0, 2)
    0.495
    >>> convolution_of_two_uniforms(-0.49, 0, 1, 0, 2)
    0.5
    >>> convolution_of_two_uniforms(0, 0, 1, 0, 2)
    0.5
    >>> convolution_of_two_uniforms(0.49, 0, 1, 0, 2)
    0.5
    >>> convolution_of_two_uniforms(0.51, 0, 1, 0, 2)
    0.495
    >>> convolution_of_two_uniforms(1.49, 0, 1, 0, 2)
    0.0050000000000000044
    >>> convolution_of_two_uniforms(1.5, 0, 1, 0, 2)
    0.0
    >>> convolution_of_two_uniforms(2, 0, 1, 0, 2)
    0.0
    """
    z = x - loc1 - loc2
    d = abs(s1 - s2)
    s = s1 + s2
    h = 2. / (d + s)

    if - s/2. <= z < - d/2.:
        x0, y0 = - s / 2., 0
        x1, y1 = -d / 2., h
        return (y1 - y0) / (x1 - x0) * (x - x0) + y0
    elif -d/2. <= z < d/2.:
        return h
    elif d/2. <= z < s/2.:
        x0, y0 = s / 2., 0
        x1, y1 = d / 2., h
        return (y1 - y0) / (x1 - x0) * (x - x0) + y0
    else:
        return 0.


def vectorized_convolution_of_two_uniforms(xs, locs1, scales1, locs2, scales2):
    return np.vectorize(convolution_of_two_uniforms)(xs, locs1, scales1, locs2, scales2)


class ConvolveTwoPdf(object):
    """
    Generate a object of probability density function, which is the convolution of two
    valid probability density function. The resulting object is able to evaluate its
    probability density at any real value.
    """

    def __init__(self, f, g, delta=1e-2, f_truncation=1e-5, g_truncation=1e-5):
        self.f = f
        self.g = g
        self.delta = delta
        self.f_truncation=f_truncation
        self.g_truncation=g_truncation

        self.model = None

    def pdf(self, x):
        if self.model is None:
            self._get_model()

        return self._pdf(x)

    def _get_model(self):
        inv_f = inversefunc(self.f, self.f_truncation)
        inv_g = inversefunc(self.g, self.g_truncation)
        assert inv_f > 0
        assert inv_g > 0
        reach = max(inv_f, inv_g)
        big_grid = np.arange(-reach, reach, self.delta)
        pmf_f = self.f(big_grid) * self.delta
        pmf_f = (pmf_f + np.hstack([pmf_f[1:], pmf_f[-1]])) / 2.  # trapezoidal rule for better accuracy
        pmf_g = self.g(big_grid) * self.delta
        pmf_g = (pmf_g + np.hstack([pmf_g[1:], pmf_g[-1]])) / 2.  # trapezoidal rule for better accuracy
        conv_pmf = scipy.signal.fftconvolve(pmf_f, pmf_g, 'same')

        # try:
        #     np.testing.assert_almost_equal(sum(conv_pmf), 1, decimal=3)
        # except AssertionError:
        #     warnings.warn('expect sum(conv_pmf) close to 1.0 but is {}'.format(sum(conv_pmf)))

        conv_pdf = conv_pmf / self.delta

        self.model = {
            'grid': big_grid,
            'pdf': conv_pdf,
        }

    def _pdf(self, x):
        assert self.model is not None
        return np.interp(x, self.model['grid'], self.model['pdf'],
                         left=self.f_truncation*self.g_truncation,
                         right=self.f_truncation*self.g_truncation)


def get_cdf(x, bins=100):
    x = np.array(x)
    counts, bin_edges = np.histogram(x, bins=bins)
    cdf = np.cumsum(counts)
    cdf = cdf / float(cdf[-1]) # normalize
    bin_edges = bin_edges[1:] # make size
    return cdf, bin_edges


def get_pdf(data, bins=20, density=True):
    pdf, bin_edges = np.histogram(data, density=density, bins=bins)
    bin_centres = (bin_edges[:-1] + bin_edges[1:])/2
    return pdf, bin_centres


def histc(series):
    dd = dict()
    for x in series:
        if np.isnan(x):
            continue
        dd.setdefault(x, 0)
        dd[x] += 1
    return dd


import os
import subprocess
import time
import unittest
from time import sleep
import multiprocessing

import numpy as np

__copyright__ = "Copyright 2016-2018, Netflix, Inc."
__license__ = "Apache, Version 2.0"


try:
    multiprocessing.set_start_method('fork')
except ValueError:  # noqa, If platform does not support, just ignore
    pass
except RuntimeError:  # noqa, If context has already being set, just ignore
    pass


def empty_object():
    return type('', (), {})()


def get_unique_sorted_list(l):
    """
    >>> get_unique_sorted_list([3, 4, 4, 1])
    [1, 3, 4]
    >>> get_unique_sorted_list([])
    []
    """
    return sorted(list(set(l)))


def get_file_name_without_extension(path):
    """

    >>> get_file_name_without_extension('yuv/src01_hrc01.yuv')
    'src01_hrc01'
    >>> get_file_name_without_extension('yuv/src01_hrc01')
    'src01_hrc01'
    >>> get_file_name_without_extension('abc/xyz/src01_hrc01.yuv')
    'src01_hrc01'
    >>> get_file_name_without_extension('abc/xyz/src01_hrc01.sdr.yuv')
    'src01_hrc01.sdr'
    >>> get_file_name_without_extension('abc/xyz/src01_hrc01.sdr.dvi.yuv')
    'src01_hrc01.sdr.dvi'

    """
    return os.path.splitext(path.split("/")[-1])[0]


def get_file_name_with_extension(path):
    """

    >>> get_file_name_with_extension('yuv/src01_hrc01.yuv')
    'src01_hrc01.yuv'
    >>> get_file_name_with_extension('src01_hrc01.yuv')
    'src01_hrc01.yuv'
    >>> get_file_name_with_extension('abc/xyz/src01_hrc01.yuv')
    'src01_hrc01.yuv'

    """
    return path.split("/")[-1]


def get_file_name_extension(path):
    '''
    >>> get_file_name_extension("file:///mnt/zli/test.txt")
    'txt'
    >>> get_file_name_extension("test.txt")
    'txt'
    >>> get_file_name_extension("abc")
    'abc'
    '''
    return path.split('.')[-1]


def indices(a, func):
    """
    Get indices of elements in an array which satisfies func
    >>> indices([1, 2, 3, 4], lambda x: x>2)
    [2, 3]
    >>> indices([1, 2, 3, 4], lambda x: x==2.5)
    []
    >>> indices([1, 2, 3, 4], lambda x: x>1 and x<=3)
    [1, 2]
    >>> indices([1, 2, 3, 4], lambda x: x in [2, 4])
    [1, 3]
    """
    return [i for (i, val) in enumerate(a) if func(val)]


def import_json_file(filepath):
    """
    Import a JSON-formatted input file as a dict.
    :param filepath:
    :return:
    """
    import json
    from argparse import Namespace
    with open(filepath, 'r') as in_f:
        ret = json.load(in_f)
    ns = Namespace(**ret)  # convert dict to namespace
    return ns


def import_python_file(filepath):
    """
    Import a python file as a module.
    :param filepath:
    :return:
    """
    filename = get_file_name_without_extension(filepath)
    try:
        from importlib.machinery import SourceFileLoader
        ret = SourceFileLoader(filename, filepath).load_module()
    except ImportError:
        import imp
        ret = imp.load_source(filename, filepath)
    return ret


def get_cmd_option(argv, begin, end, option):
    '''

    >>> get_cmd_option(['a', 'b', 'c', '--xyz', '123'], 3, 5, '--xyz')
    '123'
    >>> get_cmd_option(['a', 'b', 'c', '--xyz', '123'], 0, 5, '--xyz')
    '123'
    >>> get_cmd_option(['a', 'b', 'c', '--xyz', '123'], 4, 5, '--xyz')
    >>> get_cmd_option(['a', 'b', 'c', '--xyz', '123'], 5, 5, '--xyz')
    >>> get_cmd_option(['a', 'b', 'c', '--xyz', '123'], 6, 5, '--xyz')
    >>> get_cmd_option(['a', 'b', 'c', '--xyz', '123'], 0, 5, 'a')
    'b'
    >>> get_cmd_option(['a', 'b', 'c', '--xyz', '123'], 0, 5, 'b')
    'c'

    '''
    itr = None
    for itr in range(begin, end):
        if argv[itr] == option:
            break
    if itr is not None and itr != end and (itr + 1) != end:
        return argv[itr + 1]
    return None


def cmd_option_exists(argv, begin, end, option):
    '''

    >>> cmd_option_exists(['a', 'b', 'c', 'd'], 2, 4, 'c')
    True
    >>> cmd_option_exists(['a', 'b', 'c', 'd'], 3, 4, 'c')
    False
    >>> cmd_option_exists(['a', 'b', 'c', 'd'], 3, 4, 'd')
    True
    >>> cmd_option_exists(['a', 'b', 'c', 'd'], 2, 4, 'a')
    False
    >>> cmd_option_exists(['a', 'b', 'c', 'd'], 2, 4, 'b')
    False

    '''
    found = False
    for itr in range(begin, end):
        if argv[itr] == option:
            found = True
            break
    return found


def parallel_map(func, list_args, processes=None, pause_sec=0.01):
    """
    Build my own parallelized map function since multiprocessing's Process(),
    or Pool.map() cannot meet my both needs:
    1) be able to control the maximum number of processes in parallel
    2) be able to take in non-picklable objects as arguments
    """

    # get maximum number of active processes that can be used
    max_active_procs = processes if processes is not None else multiprocessing.cpu_count()

    # create shared dictionary
    return_dict = multiprocessing.Manager().dict()

    # define runner function
    def func_wrapper(idx_args):
        idx, args = idx_args
        executor = func(args)
        return_dict[idx] = executor

    # add idx to args
    list_idx_args = []
    for idx, args in enumerate(list_args):
        list_idx_args.append((idx, args))

    procs = []
    for idx_args in list_idx_args:
        proc = multiprocessing.Process(target=func_wrapper, args=(idx_args,))
        procs.append(proc)

    waiting_procs = set(procs)
    active_procs = set([])

    # processing
    while True:

        # check if any procs in active_procs is done; if yes, remove them
        for p in active_procs.copy():
            if not p.is_alive():
                active_procs.remove(p)

        # check if can add a proc to active_procs (add gradually one per loop)
        if len(active_procs) < max_active_procs and len(waiting_procs) > 0:
            # move one proc from waiting_procs to active_procs
            p = waiting_procs.pop()
            active_procs.add(p)
            p.start()

        # if both waiting_procs and active_procs are empty, can terminate
        if len(waiting_procs) == 0 and len(active_procs) == 0:
            break

        if pause_sec is not None:
            sleep(pause_sec) # check every x sec

    # finally, collect results
    rets = list(map(lambda idx: return_dict[idx], range(len(list_args))))

    return rets


if __name__ == '__main__':
    import doctest
    doctest.testmod()


class Timer:
    def __enter__(self):
        self.start = time.perf_counter()
        return self

    def __exit__(self, *args):
        self.end = time.perf_counter()
        self.interval = self.end - self.start


def weighed_nanmean_2d(a, weights, axis):
    assert len(a.shape) == 2
    assert axis in [0, 1]
    dim0, dim1 = a.shape
    if axis == 0:
        return np.divide(
            np.nansum(np.multiply(a,            np.tile(weights, (dim1, 1)).T), axis=0),
            np.nansum(np.multiply(~np.isnan(a), np.tile(weights, (dim1, 1)).T), axis=0)
        )
    elif axis == 1:
        return np.divide(
            np.nansum(np.multiply(a,            np.tile(weights, (dim0, 1))), axis=1),
            np.nansum(np.multiply(~np.isnan(a), np.tile(weights, (dim0, 1))), axis=1),
        )
    else:
        assert False


class MyTestCase(unittest.TestCase):

    def setUp(self):
        self.verificationErrors = []
        self.maxDiff = None

    def tearDown(self):
        self.assertEqual([], self.verificationErrors)

    def assertAlmostEqual(self, first, second, places=None, msg=None, delta=None):
        try:
            super().assertAlmostEqual(first, second, places, msg, delta)
        except AssertionError as e:
            self.verificationErrors.append(str(e))


def run_process(cmd, **kwargs):
    try:
        subprocess.check_output(cmd, stderr=subprocess.STDOUT, **kwargs)
    except subprocess.CalledProcessError as e:
        raise AssertionError(f'Process returned {e.returncode}, cmd: {cmd}, msg: {str(e.output)}')
    return 0


def cmap_factory(name):
    assert name in ['red2green', 'red2green2']
    if name == 'red2green':
        from matplotlib.colors import LinearSegmentedColormap
        cmap = LinearSegmentedColormap.from_list('rg', ["r", "w", "g"], N=256)
    elif name == 'red2green2':
        from matplotlib.colors import LinearSegmentedColormap
        c = ["darkred", "red", "lightcoral", "white",
             "palegreen", "green", "darkgreen"]
        v = [0, .15, .4, .5, 0.6, .9, 1.]
        ll = list(zip(v, c))
        cmap = LinearSegmentedColormap.from_list('rg', ll, N=256)
    else:
        assert False
    return cmap



#!/usr/bin/env python

import os
import sys
import json

from sureal.subjective_model import SubjectiveModel
from sureal.routine import run_subjective_models
from sureal.tools.misc import get_file_name_with_extension, get_cmd_option, cmd_option_exists
from sureal.config import DisplayConfig

__copyright__ = "Copyright 2016-2018, Netflix, Inc."
__license__ = "Apache, Version 2.0"

SUBJECTIVE_MODELS = ['MOS', 'MLE', 'MLE_CO', 'MLE_CO_AP', 'MLE_CO_AP2', 'DMOS', 'DMOS_MLE', 'DMOS_MLE_CO', 'SR_MOS', 'ZS_SR_MOS', 'SR_DMOS', 'ZS_SR_DMOS']


def print_usage():
    print("usage: " + os.path.basename(sys.argv[0]) + " subjective_model dataset_filepath [--output-dir output_dir]\n")
    print("subjective_model:\n\t" + "\n\t".join(SUBJECTIVE_MODELS) + "\n")


def main():
    if len(sys.argv) < 3:
        print_usage()
        return 2

    try:
        subjective_model = sys.argv[1]
        dataset_filepath = sys.argv[2]
    except ValueError:
        print_usage()
        return 2

    output_dir = get_cmd_option(sys.argv, 3, len(sys.argv), '--output-dir')
    print_ = cmd_option_exists(sys.argv, 3, len(sys.argv), '--print')

    do_plot = ['raw_scores', 'quality_scores']
    if subjective_model in ['MLE', 'MLE_CO', 'MLE_CO_AP', 'MLE_CO_AP2', 'DMOS_MLE', 'DMOS_MLE_CO']:
        do_plot.append('subject_scores')
    if subjective_model in ['MLE', 'DMOS_MLE']:
        do_plot.append('content_scores')

    try:
        subjective_model_class = SubjectiveModel.find_subclass(subjective_model)
    except Exception as e:
        print("Error: " + str(e))
        return 1

    print("Run model {} on dataset {}".format(
        subjective_model_class.__name__, get_file_name_with_extension(dataset_filepath)
    ))

    dataset, subjective_models, results = run_subjective_models(
        dataset_filepath=dataset_filepath,
        subjective_model_classes = [subjective_model_class,],
        normalize_final=False, # True or False
        do_plot=do_plot,
        plot_type='errorbar',
        gradient_method='simplified',
    )

    if print_:
        print(("Dataset: {}".format(dataset_filepath)))
        print(("Subjective Model: {} {}".format(subjective_models[0].TYPE, subjective_models[0].VERSION)))
        print("Result:")
        printable_results = {k: list(v) for k, v in results[0].items() if isinstance(v, list)}
        print(json.dumps(printable_results, indent=4, sort_keys=True))

    if output_dir is None:
        DisplayConfig.show()
    else:
        print(("Output wrote to {}.".format(output_dir)))
        if not os.path.exists(output_dir):
            os.makedirs(output_dir)
        DisplayConfig.show(write_to_dir=output_dir)
        with open(os.path.join(output_dir, 'sureal.json'), 'w') as out_f:
            json.dump(results[0], out_f, default=lambda o: '<not serializable>',
                      indent=4, sort_keys=True)
    return 0


if __name__ == '__main__':
    ret = main()
    exit(ret)



__author__ = "Alvaro Sanchez-Gonzalez"
__license__ = "MIT"

import numpy as np
import warnings

from scipy.optimize import minimize_scalar

__all__ = ['inversefunc']


def inversefunc(func,
                y_values=None,
                domain=None,
                image=None,
                open_domain=None,
                args=(),
                accuracy=2):
    r"""Obtain the inverse of a function.

    Returns the numerical inverse of the function `f`. It may return a callable
    that can be used to calculate the inverse, or the inverse of certain points
    depending on the `y_values` argument.
    In order for the numerical inverse to exist in its domain, the
    input function must have, continuous, strictly monotonic behavior i.e. be
    purely decreasing or purely increasing in that domain. By default the
    domain interval spans all the real numbers, however it can be restricted
    with the `domain` and `open_domain` arguments. The image of the function
    in the interval may be provided, for cases where the function is non
    continuous right at the end of an open interval.

    Parameters
    ----------
    func : callable
        Callable representing the function to be inverted, able to take a
        ndarray or an scalar and return an object of the same kind with the
        evaluation of the function.  If `func` takes many arguments, it is
        inverted along the axis corresponding to the first argument.
        The function must not diverge and have a continuous strictly monotonic
        behavior in the chosen interval.
    y_values : float, ndarray, optional
        Values for which calculate the inverse function. If set to None, then
        a callable that can be used to calculate the inverse of values is
        returned. Default None.
    domain : float, ndarray, optional
        Boundaries of the domain (`domain[0]`, `domain[1]`).
        `domain[1]` must be larger than `domain[0]`.
        None values are assumed to be no boundary in that direction.
        A single scalar value will set it to [`domain`, None].
        Default None (-Inf, Inf).
    open_domain : bool, ndarray, optional
        Whether the domain is an open interval at each of the ends.
        A single scalar boolean will set it to [`open_domain`, `open_domain`].
        Default None [False, False].
    image : float, ndarray, optional
        Image of the function in the domain (`image[0]`, `image[1]`).
        `image[1]` must be larger than `image[0]`.
        None values are assumed to be no boundary in that direction.
        Default None, this is (-Inf, Inf) if domain is None, or the limits
        set by func(domain[0]) and func(domain[1]).
    args : tuple, optional
        Extra arguments to pass to `func`. Default ().
    accuracy : int, optional
        Number of digits for the desired accuracy. It will give a warning
        if the accuracy is worse than this.
        Default 2.

    Returns
    -------
    callable or ndarray
        Inverse function of `func`. It can take scalars or ndarrays, and return
        objects of the same kind with the calculated inverse values.

    Notes
    -----

    .. versionadded:: 0.19.0

    Examples
    --------
    >>> import numpy as np
    >>> cube = (lambda x: x**3)
    >>> invcube = inversefunc(cube)
    >>> invcube(27) # Should give 3
    array(3.00000001)
    >>> invsquare = inversefunc(np.power, args=(2), domain=0)
    >>> invsquare([4, 16, 64]) # Should give [2, 4, 8]
    array([2., 4., 8.])
    >>> inversefunc(np.log10, y_values=-2, # Should give 0.01
    ...             domain=0, open_domain=True)
    array(0.01)
    >>> inversefunc(np.cos, y_values=[1, 0, -1], # Should give [0, pi / 2, pi]
    ...             domain=[0, np.pi])
    array([0.        , 1.57079632, 3.14159265])
    >>> invtan = inversefunc(np.tan,
    ...                      domain=[-np.pi / 2, np.pi / 2],
    ...                      open_domain=True)
    >>> invtan([1, 0, -1]) # Should give [pi / 4, 0, -pi / 4]
    array([ 7.85398163e-01,  1.29246971e-26, -7.85398163e-01])

    """

    domain, image, open_domain, args = _normparams_inversefunc(domain,
                                                               image,
                                                               open_domain,
                                                               args)

    ymin, ymax = image
    xmin, xmax = domain
    xmin_open, xmax_open = open_domain

    # Calculating if the function is increasing or decreasing, using ref points
    # anywhere in the valid range (Function has to be strictly monotonic)
    ref1, ref2 = _get_valid_refpoints(xmin, xmax)
    trend = np.sign(func(ref2, *args) - func(ref1, *args))

    if trend == 0:
        raise ValueError("Function is not strictly monotonic")

    # Calculating the image by default
    if ymin is None:
        ymin = _auto_ymin(func, args, xmin, xmax, trend)
    if ymax is None:
        ymax = _auto_ymax(func, args, xmin, xmax, trend)

    # Creating bounded function
    def bounded_f(x):
        if xmin is not None and (x < xmin or (x == xmin and xmin_open)):
                val = -1 * np.inf * trend
        elif xmax is not None and (x > xmax or (x == xmax and xmax_open)):
                val = np.inf * trend
        else:
            val = func(x, *args)
        return val

    min_kwargs = {}
    min_kwargs['bracket'] = (ref1, ref2)
    min_kwargs['tol'] = 1.48e-08
    min_kwargs['method'] = 'Brent'

    def inv(yin):
        yin = np.asarray(yin, dtype=np.float64)
        shapein = yin.shape
        yin = yin.flatten()
        if ymin is not None:
            if (xmin_open and trend == 1) or (xmax_open and trend == -1):
                mask = yin <= ymin
            else:
                mask = yin < ymin
            if yin[mask].size > 0:
                raise ValueError("Requested values %s lower than the"
                                 " lower limit %g of the image" %
                                 (yin[mask], ymin))
        if ymax is not None:
            if (xmax_open and trend == 1) or (xmin_open and trend == -1):
                mask = yin >= ymax
            else:
                mask = yin > ymax
            if yin[mask].size > 0:
                raise ValueError("Requested values %s higher than the"
                                 " higher limit %g of the image" %
                                 (yin[mask], ymax))

        results = yin.copy() * np.nan
        resultsmask = np.zeros(yin.shape, dtype=bool)

        for j in range(yin.size):
            if xmax is not None:
                if bounded_f(xmax) == yin[j]:
                    results[j] = xmax
                    resultsmask[j] = True
                    continue
            if xmin is not None:
                if bounded_f(xmin) == yin[j]:
                    results[j] = xmin
                    resultsmask[j] = True
                    continue

            optimizer = (lambda x, j=j,
                         bounded_f=bounded_f: (((bounded_f(x) - yin[j]))**2))
            try:
                with warnings.catch_warnings(record=True):
                    result = minimize_scalar(optimizer, **min_kwargs)
                results[j] = result.x
                resultsmask[j] = result.success
            except:
                resultsmask[j] = False
        if any(~resultsmask):
            warnings.warn("Trouble calculating inverse for values: "
                          "%s" % str(yin[~resultsmask]), RuntimeWarning)

        # try:
        #     np.testing.assert_array_almost_equal(yin, func(results, *args),
        #                                          decimal=accuracy)
        # except AssertionError:
        #     warnings.warn("Results obtained with less than %g "
        #                   "decimal digits of accuracy"
        #                   % accuracy, RuntimeWarning)

        return results.reshape(shapein)

    if y_values is None:
        return inv
    else:
        return inv(y_values)


def _normparams_inversefunc(domain, image, open_domain, args):

    if not isinstance(args, tuple):
        args = (args,)

    error_domain = ("domain must be a single scalar, or a have two "
                    "elements [xmin, xmax]. Set None, to leave it "
                    "unlimited on one side.")
    if domain is None:
        domain = (None, None)
    else:
        domain = np.asarray(domain)
        if domain.ndim == 0:
            domain = (float(domain), None)
        elif domain.ndim == 1 and domain.size != 2:
            raise ValueError(error_domain)
        elif domain.ndim > 1:
            raise ValueError(error_domain)
        else:
            domain = ((float(domain[0]) if domain[0] is not None else None),
                      (float(domain[1]) if domain[1] is not None else None))

    error_open_domain = ("open_domain must be a single scalar, or a have two "
                         "bool elements [open_xmin, open_xmax].")
    if open_domain is None:
        open_domain = (False, False)
    else:
        open_domain = np.asarray(open_domain)
        if open_domain.ndim == 0:
            open_domain = (bool(open_domain), bool(open_domain))
        elif open_domain.ndim == 1 and open_domain.size != 2:
            raise ValueError(error_open_domain)
        elif open_domain.ndim > 1:
            raise ValueError(error_open_domain)
        else:
            open_domain = (bool(open_domain[0]), bool(open_domain[1]))

    error_image = ("image must be a single scalar, or a have two "
                   "bool elements [ymin, ymax].")
    if image is None:
        image = (None, None)
    else:
        image = np.asarray(image)
        if image.ndim != 1 or image.size != 2:
            raise ValueError(error_image)
        else:
            image = ((float(image[0]) if image[0] is not None else None),
                     (float(image[1]) if image[1] is not None else None))

    if domain[0] is not None and domain[1] is not None:
        if domain[0] >= domain[1]:
            raise ValueError("domain[0] min must be less than domain[1]")

    if image[0] is not None and image[1] is not None:
        if image[0] >= image[1]:
            raise ValueError("image[0] min must be less than image[1]")

    return domain, image, open_domain, args


def _get_valid_refpoints(xmin, xmax):
    if xmin is not None and xmax is not None:
        d = xmax - xmin
        ref1 = xmin + d / 4.
        ref2 = xmax - d / 4.
    elif xmin is not None:
        ref1 = xmin + 1.
        ref2 = xmin + 2.
    elif xmax is not None:
        ref1 = xmax - 2.
        ref2 = xmax - 1.
    else:
        ref1 = 0.
        ref2 = 1.
    return ref1, ref2


def _auto_ymin(func, args, xmin, xmax, trend):
    ymin = None
    if ((xmin is not None and trend == 1) or
       (xmax is not None and trend == -1)):
        try:
            with warnings.catch_warnings(record=True):
                ymin = func(xmin, *args) if trend == 1 else func(xmax, *args)
        except:
            raise ValueError("Cannot automatically calculate the lower limit "
                             "of the image please inclue it as a parameter")
    return ymin


def _auto_ymax(func, args, xmin, xmax, trend):
    ymax = None
    if ((xmax is not None and trend == 1) or
       (xmin is not None and trend == -1)):
        try:
            with warnings.catch_warnings(record=True):
                ymax = func(xmax, *args) if trend == 1 else func(xmin, *args)
        except:
            raise ValueError("Cannot automatically calculate the upper limit "
                             "of the image please include it as a parameter")
    return ymax




from abc import abstractmethod, ABCMeta

import re

__copyright__ = "Copyright 2016-2018, Netflix, Inc."
__license__ = "Apache, Version 2.0"


class TypeVersionEnabled(object):
    """
    Mandate a type name and a version string. Derived class (e.g. an Executor)
    has a unique string combining type and version. The string is useful in
    identifying a Result by which Executor it is generated (e.g. VMAF_V0.1,
    PSNR_V1.0, or VMAF_feature_V0.1).
    """

    __metaclass__ = ABCMeta

    @property
    @abstractmethod
    def TYPE(self):
        raise NotImplementedError

    @property
    @abstractmethod
    def VERSION(self):
        raise NotImplementedError

    def __init__(self):
        self._assert_type_version()

    def _assert_type_version(self):

        # assert re.match(r"^[a-zA-Z0-9_]+$", self.TYPE), \
        #     "TYPE can only contains alphabets, numbers and underscore (_)."

        assert re.match(r"^[a-zA-Z0-9._-]+$", self.VERSION), \
            "VERSION can only contains alphabets, numbers, dot (.), hyphen(-) and underscore (_)."

    def get_type_version_string(self):
        return "{type}_V{version}".format(type=self.TYPE,
                                          version=self.VERSION)

    def get_cozy_type_version_string(self):
        return "{type} VERSION {version}".format(type=self.TYPE,
                                                 version=self.VERSION)

    @classmethod
    def find_subclass(cls, subclass_type):
        """
        Find subclass by TYPE.
        :param subclass_type:
        :return:
        """
        matched_subclasses = []
        for subclass in cls.get_subclasses_recursively():
            if hasattr(subclass, 'TYPE') and subclass.TYPE == subclass_type:
                matched_subclasses.append(subclass)
        assert len(matched_subclasses) == 1, \
            "Must have one and only one subclass of {class_name} with type " \
            "{type}, but got {num}".format(
                class_name=cls.__name__,
                type=subclass_type,
                num=len(matched_subclasses))
        return matched_subclasses[0]

    @classmethod
    def get_subclasses_recursively(cls):
        subclasses = cls.__subclasses__()
        subsubclasses = []
        for subclass in subclasses:
            subsubclasses += subclass.get_subclasses_recursively()
        return subclasses + subsubclasses




#!/usr/bin/env python3

import argparse
import json
import os

from sureal.config import DisplayConfig
from sureal.dataset_reader import RawDatasetReader, PairedCompDatasetReader
from sureal.pc_subjective_model import PairedCompSubjectiveModel
from sureal.routine import run_subjective_models, \
    format_output_of_run_subjective_models
from sureal.subjective_model import SubjectiveModel


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--dataset", dest="dataset", nargs=1, type=str,
        help="Path to the dataset file.",
        required=True)
    parser.add_argument(
        "--models", dest="models", nargs="+", type=str,
        help="Subjective models to use (can specify more than one), choosing "
             "from: MOS, P910, P913, BT500.",
        required=True)
    parser.add_argument(
        "--output-dir", dest="output_dir", nargs=1, type=str,
        help="Path to the output directory (will force create is not existed). "
             "If not specified, plots will be displayed and output will be printed.",
        required=False)
    parser.add_argument(
        "--plot-raw-data", dest="plot_raw_data", action='store_true',
        help="Plot the raw data. This includes the raw opinion scores presented "
             "in a video-subject matrix, counts per video and counts per subject.",
        required=False)
    parser.add_argument(
        "--plot-dis-videos", dest="plot_dis_videos", action='store_true',
        help="Plot the subjective scores of the distorted videos.",
        required=False)
    parser.add_argument(
        "--plot-observers", dest="plot_observers", action='store_true',
        help="Plot the scores of the observers.",
        required=False)
    args = parser.parse_args()
    dataset = args.dataset[0]
    models = args.models
    output_dir = args.output_dir[0] if args.output_dir else None
    plot_raw_data = args.plot_raw_data
    plot_dis_videos = args.plot_dis_videos
    plot_observers = args.plot_observers

    ModelClasses = list()
    for model in models:
        ModelClass = SubjectiveModel.find_subclass(model)
        ModelClasses.append(ModelClass)

    def is_subj_model_class(ModelClass):
        superclasses = ModelClass.__mro__
        return SubjectiveModel in superclasses and PairedCompSubjectiveModel not in superclasses

    def is_pc_subj_model_class(ModelClass):
        return PairedCompSubjectiveModel in ModelClass.__mro__

    # ModelClass should be either SubjectiveModel or PairedCompSubjectiveModel
    is_all_subjective_model = all([is_subj_model_class(ModelClass) for ModelClass in ModelClasses])
    is_all_pc_subjective_model = all([is_pc_subj_model_class(ModelClass) for ModelClass in ModelClasses])
    assert (is_all_subjective_model and not is_all_pc_subjective_model) or \
           (not is_all_subjective_model and is_all_pc_subjective_model), \
        f'is_all_subjective_model: {is_all_subjective_model}, ' \
        f'is_all_pc_subjective_model: {is_all_pc_subjective_model}'
    if is_all_subjective_model:
        DatasetReaderClass = RawDatasetReader
    else:
        DatasetReaderClass = PairedCompDatasetReader

    do_plot = []
    if plot_raw_data:
        do_plot.append('raw_scores')
        if is_all_subjective_model:
            do_plot.append('raw_counts')
            do_plot.append('raw_counts_per_subject')
    if plot_dis_videos:
        do_plot.append('quality_scores')
    if plot_observers:
        do_plot.append('subject_scores')

    dataset, subjective_models, results = run_subjective_models(
        dataset_filepath=dataset,
        subjective_model_classes=ModelClasses,
        do_plot=do_plot,
        dataset_reader_class=DatasetReaderClass,
    )

    output = format_output_of_run_subjective_models(
        dataset, subjective_models, results)

    if output_dir is None:
        json_output = json.dumps(output, indent=4)
        print(json_output)
        DisplayConfig.show()
    else:
        os.makedirs(output_dir, exist_ok=True)
        with open(os.path.join(output_dir, 'output.json'), 'w') as fp:
            json.dump(output, fp, indent=4)
        DisplayConfig.show(write_to_dir=output_dir)
        print(f'output is written to directory {output_dir}.')

    return 0


if __name__ == '__main__':
    ret = main()
    exit(ret)
